#!/usr/bin/env python3
"""
This module provides testing code for the
auxi.modelling.business.activity module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
import collections
from datetime import datetime
from dateutil.relativedelta import relativedelta

from auxi.core.time import Clock
from auxi.modelling.business.structure import Activity
from auxi.modelling.business.structure import Component
from auxi.modelling.business.structure import Entity
from auxi.modelling.business.basic import BasicActivity
from auxi.modelling.financial.des import GeneralLedger
from auxi.modelling.financial.des import GeneralLedgerStructure


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ActivityUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.business.structure.Activity class.
    """

    def setUp(self):
        self.object = Activity("NameA",
                               description="DescriptionA",
                               start=datetime(2016, 2, 1),
                               end=datetime(2017, 2, 1),
                               interval=3)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.end_datetime, datetime(2017, 2, 1))
        self.assertEqual(self.object.interval, 3)

    def test_get_path(self):
        self.object.set_parent_path("entityA/componentA")
        self.assertEqual(self.object.path, "entityA/componentA/NameA")

    def test_set_parent_path(self):
        self.object.set_parent_path("entityA/componentA")
        self.assertEqual(self.object.path, "entityA/componentA/NameA")

    def test_set_name(self):
        """
        Test wether the name changes when it is set
        and that the activity's path is updated correctly.
        """
        self.object.set_parent_path("entityA/componentA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.path, "entityA/componentA/NameAt")

    def test__meet_exection_criteria(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 13)
        self.assertEqual(self.object._meet_execution_criteria(3), False)
        self.assertEqual(self.object._meet_execution_criteria(5), True)
        self.assertEqual(self.object._meet_execution_criteria(40), False)

    def test_prepare_to_run(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 18)
        self.assertEqual(self.object.start_period_ix, 2)
        self.assertEqual(self.object.end_period_ix, 14)

    def test_get_referenced_accounts(self):
        self.assertEqual(len(self.object.get_referenced_accouts()), 0)


class ComponentUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.business.structure.Component class.
    """

    def setUp(self):
        self.gl_structure = GeneralLedgerStructure(
            "NameA",
            description="DescriptionA")
        self.gl_structure["Sales"].create_account("Default", number="0000")
        self.gl = GeneralLedger(
            "NameA",
            self.gl_structure,
            description="DescriptionA")
        self.object = Component("ComponentA",
                                self.gl,
                                description="DescriptionA")
        # Set up the needed objects
        self.object.create_component("ComponentA1", description="ca1")
        self.object.create_component("ComponentA2", description="ca2")
        self.basic_activity = BasicActivity(
            "BasicActivityA",
            description="DescriptionA",
            dt_account="Bank/Default",
            cr_account="Sales/Default",
            amount=5000,
            start=datetime(2016, 2, 1),
            end=datetime(2017, 2, 1),
            interval=3)
        self.object["ComponentA1"].add_activity(self.basic_activity)

    def test_constructor(self):
        self.assertEqual(self.object.name, "ComponentA")
        self.assertEqual(self.object.description, "DescriptionA")

    def test_set_parent_path(self):
        self.object.set_parent_path("/entityA")
        self.assertEqual(self.object.path, "/entityA/ComponentA")

    def test_set_name(self):
        """
        Test wether the name changes when it is set, that the component's
        name changes and that the component's children's paths are updated
        correctly.
        """
        self.object.set_parent_path("entityA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.components[0].path,
                         "entityA/NameAt/ComponentA1")
        self.assertEqual(self.object.components[1].path,
                         "entityA/NameAt/ComponentA2")
        self.assertEqual(self.object.components[0].activities[0].path,
                         "entityA/NameAt/ComponentA1/BasicActivityA")

    def test_create_component(self):
        new_comp = self.object.create_component("ComponentTestCreate",
                                                description="catc")
        self.assertEqual(new_comp.name, "ComponentTestCreate")
        self.assertEqual(new_comp.description, "catc")

        self.assertEqual(new_comp, self.object.components[2])

    def test_remove_component(self):
        self.object.create_component("ComponentTestRemove",
                                     description="catr")
        self.object.remove_component("ComponentTestRemove")
        self.assertEqual(len(self.object.components), 2)

    def test_get_component_exists(self):
        self.assertEqual(
            self.object.get_component("ComponentA1"),
            self.object["ComponentA1"])

    def test_get_component_not_exists(self):
        self.assertRaises(IndexError, self.object.get_component, "ComponentA3")

    def test_add_activity_valid_account_name(self):
        basic_activity2 = BasicActivity(
            "BasicActivityB",
            description="DescriptionB",
            dt_account="Bank/Default",
            cr_account="Sales/Default",
            amount=5000,
            start=datetime(2016, 2, 1),
            end=datetime(2017, 2, 1),
            interval=3)
        self.object["ComponentA2"].add_activity(basic_activity2)

        self.assertEqual(len(self.object["ComponentA2"].activities), 1)
        self.assertEqual(
            basic_activity2.path,
            "/ComponentA/ComponentA2/BasicActivityB")

    def test_add_activity_invalid_account_name(self):
        basic_activity2 = BasicActivity(
            "BasicActivityB",
            description="DescriptionB",
            dt_account="invalid_acc_name_a",
            cr_account="invalid_acc_name_b",
            amount=5000,
            start=datetime(2016, 2, 1),
            end=datetime(2017, 2, 1),
            interval=3)
        self.assertRaises(
            ValueError,
            self.object["ComponentA2"].add_activity, basic_activity2)

    def test_get_activity_exists(self):
        self.assertEqual(
            self.object["ComponentA1"].get_activity("BasicActivityA"),
            self.basic_activity)

    def test_get_activity_not_exists(self):
        self.assertRaises(
            IndexError,
            self.object["ComponentA1"].get_activity, "B")

    def test_prepare_to_run(self):
        """
        Test that the component run's its activities' prepare_to_run methods.
        """
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 18)
        # The activity's start and end period indexes should be changed.
        self.assertEqual(
            self.object.components[0].activities[0].start_period_ix, 2)
        self.assertEqual(
            self.object.components[0].activities[0].end_period_ix, 14)

    def test_run(self):
        """
        Test that the component runs its activities.
        """
        # Set up the general ledger so that an activity can create a
        # transaction.
        structure = GeneralLedgerStructure("NameA", description="DescriptionA")
        gl = GeneralLedger("NameA", structure, description="DescriptionA")
        # Prepare the component for a run. Tick the clock to the correct place
        # for the activity to create a transaction.
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        clock.tick()
        clock.tick()
        self.object.prepare_to_run(clock, 20)
        self.object.run(clock, gl)
        # If the activity has been run, then a transaction should have
        # been generated.
        self.assertEqual(len(gl.transactions), 1)


class EntityUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.entity.component class.
    """

    def setUp(self):
        self.gl_structure = GeneralLedgerStructure(
            "NameA",
            description="DescriptionA")
        self.sales_acc = self.gl_structure._acci_sales_.create_account(
            "sales_default",
            "0000")
        self.cos_acc = self.gl_structure._acci_cos_.create_account(
            "cos_default",
            "0000")
        self.reva_acc = self.gl_structure["Other Income"].create_account(
            "RevA",
            number="011")
        self.expa_acc = self.gl_structure["Expense"].create_account(
            "ExpenseA",
            number="011")

        self.object = Entity("EntityA",
                             gl_structure=self.gl_structure,
                             description="DescriptionA")
        # Set up the needed objects
        self.comp1 = self.object.create_component("ComponentA1",
                                                  description="ca1")
        basic_activity = BasicActivity("BasicActivityA",
                                       description="DescriptionA",
                                       dt_account="Bank",
                                       cr_account="Sales",
                                       amount=5000,
                                       start=datetime(2016, 2, 1),
                                       end=datetime(2017, 2, 1),
                                       interval=1)
        self.comp1.add_activity(basic_activity)

        self.clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))

    def test_constructor(self):
        self.assertEqual(self.object.name, "EntityA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.gl.structure.name, self.gl_structure.name)
        self.assertEqual(self.object.period_count, -1)

    def test_set_parent_path(self):
        self.object.set_parent_path("/modelA")
        self.assertEqual(self.object.path, "/modelA/EntityA")

    def test_set_name(self):
        """
        Test wether the name changes when it is set, that the entity's
        name changes and that the component's children's paths are updated
        correctly.
        """

        self.object.set_parent_path("modelA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.components[0].path,
                         "modelA/NameAt/ComponentA1")
        self.assertEqual(self.object.components[0].activities[0].path,
                         "modelA/NameAt/ComponentA1/BasicActivityA")

    def test_getitem_exists(self):
        self.assertEqual(
            self.comp1,
            self.object["ComponentA1"])

    def test_getitem_not_exists(self):
        self.assertRaises(IndexError, self.object.__getitem__, "ComponentA3")

    def test_create_component(self):
        new_comp = self.object.create_component("ComponentA2",
                                                description="ca2")
        self.assertEqual(new_comp.name, "ComponentA2")
        self.assertEqual(new_comp.description, "ca2")

        self.assertEqual(new_comp, self.object.components[1])

    def test_remove_component(self):
        self.object.create_component("ComponentA2",
                                     description="ca2")
        self.object.remove_component("ComponentA2")
        self.assertEqual(len(self.object.components), 1)

    def test_prepare_to_run(self):
        """
        Test that the entity run's its component' prepare_to_run methods.
        """

        # Test the prepare to run method before any run has been performed
        self.object.gl.create_transaction(
            "TestA",
            description="TestA_Desc",
            tx_date=datetime(2016, 2, 1).date(),
            dt_account="Bank",
            cr_account="Sales",
            source="Peanut Sales",
            amount=20.00)
        self.object.negative_income_tax_total = 99

        self.object.prepare_to_run(self.clock, 18)

        # Was the component's prepare to run executed?
        self.assertEqual(
            self.object.components[0].activities[0].start_period_ix, 2)
        self.assertEqual(
            self.object.components[0].activities[0].end_period_ix, 14)
        # Was the datetime's set correctly?
        self.assertEqual(
            self.object._exec_year_end_datetime,
            self.clock.get_datetime_at_period_ix(18))
        self.assertEqual(
            self.object._prev_year_end_datetime,
            self.clock.start_datetime)
        self.assertEqual(
            self.object._curr_year_end_datetime,
            self.clock.start_datetime + relativedelta(years=1))
        # Was the transactions cleared?
        self.assertEqual(len(self.object.gl.transactions), 0)
        # Was the negative tax income reset to 0?
        self.assertEqual(self.object.negative_income_tax_total, 0)

    def test_run(self):
        """
        Test that the entity runs its components.
        """

        self.object.prepare_to_run(self.clock, 20)
        self.clock.tick()
        self.clock.tick()
        self.object.run(self.clock)
        self.assertEqual(len(self.object.gl.transactions), 1)
        self.assertEqual(self.object.gl.transactions[0].amount, 5000)

    def test__perform_year_end_gross_profit_and_income_summary(self):
        """
        Test that the year end gross profit and income summary accounts
        are summed up correctly.
        """

        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        # Create the accounts to write of dict
        income_summary_write_off_accounts = collections.OrderedDict([
            (self.cos_acc, 700),
            (gls._acci_inc_tax_exp_acc_, -300),
            (self.sales_acc, 70),
            (gls._accb_bank_["Default"], -30)])
        gross_profit_write_off_accounts = collections.OrderedDict([
            (self.sales_acc, 70),
            (gls._accb_bank_["Default"], -30)])
        # Test when there the gross profit is 0.
        # The gross profit account should not be settled:
        inc = self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            {},
            income_summary_write_off_accounts)
        # The gross profit account should not be settled.
        # Only the 4 transactions from the income summary year end calcs
        # should be there
        self.assertEqual(len(self.object.gl.transactions), 4)
        # The income summary should be that of the income summary year end calc
        # method: -360 (See the income summary year end calc, in this case,
        # gross profit starts at 0 instead of 5000, thus -360 instead of 4640)
        self.assertEqual(inc, -360)

        # Test when there is gross profit.
        # The gross profit account should not be settled:
        del self.object.gl.transactions[:]  # Clear the tx list
        self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            gross_profit_write_off_accounts, {})
        # The gross profit account should be settled.
        # Only the 2 transactions from the income summary year end calcs
        # should be there as well as the gross profit settle transaction
        self.assertEqual(len(self.object.gl.transactions), 3)
        # Make sure the gross profit settle transaction appears correct
        # When gross profit is positive:
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[2].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 40)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)
        # When gross profit is negative:
        gross_profit_write_off_accounts = collections.OrderedDict([
            (self.sales_acc, -70),
            (gls._accb_bank_["Default"], 30)])
        del self.object.gl.transactions[:]  # Clear the tx list
        self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            gross_profit_write_off_accounts, {})
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[2].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 40)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)

    def test__perform_year_end_gross_profit(self):
        """
        Test that the year end gross profit account is closed correctly.
        """

        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)

        # Create the accounts to write of dict
        gross_profit_write_off_accounts = collections.OrderedDict([
            (self.sales_acc, 70),
            (gls._accb_bank_["Default"], -30)])
        gross_profit = self.object._perform_year_end_gross_profit(
            year_end_datetime, gross_profit_write_off_accounts)
        self.assertEqual(len(self.object.gl.transactions), 2)
        # Expenses: 700-300 = 400
        # Other: 70-30 = 40
        # Total: gross_profit + Other - expenses: 5000 + 40 - 400 = 4640
        self.assertEqual(gross_profit, 40)

        # Test the "self.sales_acc: 70" transaction
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         self.sales_acc.path)
        self.assertEqual(self.object.gl.transactions[0].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount, 70)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test the "gls._accb_bank_["Default"]: -30" transaction
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls._accb_bank_["Default"].path)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[1].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount, 30)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

    def test__perform_year_end_income_summary(self):
        """
        Test that the year end income summary account is closed correctly.
        """

        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        gross_profit = 5000

        # Create the accounts to write of dict
        income_summary_write_off_accounts = collections.OrderedDict([
            (self.cos_acc, 700),
            (gls._acci_inc_tax_exp_acc_, -300),
            (self.sales_acc, 70),
            (gls._accb_bank_["Default"], -30)])
        income_summary_amount = self.object._perform_year_end_income_summary(
            year_end_datetime, gross_profit, income_summary_write_off_accounts)
        self.assertEqual(len(self.object.gl.transactions), 4)
        # Expenses: 700-300 = 400
        # Other: 70-30 = 40
        # Total: gross_profit + Other - expenses: 5000 + 40 - 400 = 4640
        self.assertEqual(income_summary_amount, 4640)

        # Test the "self.cos_acc: 700" transaction
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         self.cos_acc.path)
        self.assertEqual(self.object.gl.transactions[0].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount, 700)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test the "gls._acci_inc_tax_exp_acc_: -300" transaction
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls._acci_inc_tax_exp_acc_.path)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[1].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount, 300)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

        # Test the "self.sales_acc: 70" transaction
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         self.sales_acc.path)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[2].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 70)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)

        # Test the "gls._accb_bank_["Default"]: -30" transaction
        self.assertEqual(self.object.gl.transactions[3].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[3].cr_account,
                         gls._accb_bank_["Default"].path)
        self.assertEqual(self.object.gl.transactions[3].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[3].amount, 30)
        self.assertEqual(self.object.gl.transactions[3].is_closing_cr_account,
                         True)

    def test__perform_year_end_income_tax(self):
        # TODO
        pass

    def test__perform_year_end_retained_earnings(self):
        """
        Test that the year end retained earnings account is closed correctly.
        """

        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        # Test for == 0 case
        income_summary = 0
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 0)

        # Test for > 0 case
        income_summary = 1000
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 1)
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         gls._accb_ret_earnings_acc_.path)
        self.assertEqual(self.object.gl.transactions[0].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount,
                         income_summary)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test for < 0 case
        income_summary = -300
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 2)
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls._accb_ret_earnings_acc_.path)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[1].tx_date,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount,
                         income_summary)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

    def test__perform_year_end_procedure(self):
        """
        Test that all the year end accounts has been closed of correctly.
        """

        gls = self.object.gl.structure
        # Test when that no year end transactions are created in the
        # middle of a year
        self.object.prepare_to_run(self.clock, 28)
        self.clock.timestep_ix = 6
        # create the transactions
        self.object.gl.create_transaction(
            "Sales_tx",
            description="Sales tx",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls._accb_bank_["Default"].path,  # Bank account
            cr_account=self.sales_acc.path,
            source=self.object.path,
            amount=70)
        # create the transactions
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx",
            description="Cost of Sales tx",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.cos_acc.path,
            cr_account=gls._accb_bank_["Default"].path,  # Bank account
            source=self.object.path,
            amount=30)
        self.object._perform_year_end_procedure(self.clock)
        # Only one transaction should have been created and no
        # year end transactions.
        self.assertEqual(len(self.object.gl.transactions), 2)

        # Test that the year end procedure, at the year end, creates
        #   - two gross profit transactions with values of 70 and 40
        #   - one gross profit and income summary settle transaction
        #      with a value of 70 - 30
        #   - one retained earnings transaction with a value of 70 - 30
        # Only the 2 transaction (sales tx, and cost of sales tx) and the
        # 4 'year end' transactions should have been created.)
        self.clock.timestep_ix = 12
        self.object._perform_year_end_procedure(self.clock)

        self.assertEqual(len(self.object.gl.transactions), 6)
        self.assertEqual(self.object.gl.transactions[2].amount, 70)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         self.sales_acc.path)

        self.assertEqual(self.object.gl.transactions[3].amount, 30)
        self.assertEqual(self.object.gl.transactions[3].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[3].dt_account,
                         self.cos_acc.path)
        self.assertEqual(self.object.gl.transactions[3].cr_account,
                         gls._acci_gross_prof_.path)

        self.assertEqual(self.object.gl.transactions[4].amount, 40)
        self.assertEqual(self.object.gl.transactions[4].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[4].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[4].cr_account,
                         gls._acci_inc_sum_.path)

        self.assertEqual(self.object.gl.transactions[5].amount, 40)
        self.assertEqual(self.object.gl.transactions[5].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[5].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[5].cr_account,
                         gls._accb_ret_earnings_acc_.path)

        # Test for a new year
        self.clock.timestep_ix = 18
        # Test for when other revenue and expense accounts transactions
        # exists as well.

        # create a sales transactions
        self.object.gl.create_transaction(
            "SalesTx2",
            description="Sales tx2",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls._accb_bank_["Default"].path,  # Bank account
            cr_account=self.sales_acc.path,
            source=self.object.path,
            amount=45)
        self.object.gl.create_transaction(
            "SalesTx3",
            description="Sales tx3",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls._accb_bank_["Default"].path,  # Bank account
            cr_account=self.sales_acc.path,
            source=self.object.path,
            amount=35)
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx2",
            description="Cost of Sales tx 2",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.cos_acc.path,
            cr_account=gls._accb_bank_["Default"].path,  # Bank account
            source=self.object.path,
            amount=15)
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx3",
            description="Cost of Sales tx 3",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.cos_acc.path,
            cr_account=gls._accb_bank_["Default"].path,  # Bank account
            source=self.object.path,
            amount=5)
        # create the transactions
        self.object.gl.create_transaction(
            "revenueATx",
            description="revenue a tx",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls._accb_bank_["Default"].path,  # Bank account
            cr_account=self.reva_acc.path,
            source=self.object.path,
            amount=450)
        self.object.gl.create_transaction(
            "revenueATx2",
            description="revenue a tx2",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls._accb_bank_["Default"].path,  # Bank account
            cr_account=self.reva_acc.path,
            source=self.object.path,
            amount=350)
        # create the transactions
        self.object.gl.create_transaction(
            "expenseAtx",
            description="expense a tx",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.expa_acc.path,
            cr_account=gls._accb_bank_["Default"].path,  # Bank account
            source=self.object.path,
            amount=320)
        self.object.gl.create_transaction(
            "expenseAtx2",
            description="expense a tx2",
            tx_date=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.expa_acc.path,
            cr_account=gls._accb_bank_["Default"].path,  # Bank account
            source=self.object.path,
            amount=180)

        # Test that the year end procedure, at the year end, creates
        #   - two gross profit transactions with values of 80 and 20
        #   - two income summary transactions with values of 800 and 500
        #   - one gross profit and income summary settle transaction
        #      with a value of 80 - 20
        #   - one retained earnings transaction with a value of
        #      800 - 500 + 80 - 20
        # Only the 6 transactions (2 sales tx, 2 cost of sales,
        # 2 revenue a and 2 expense a) and the 5 'year end' transactions
        # should have been created.)
        self.clock.timestep_ix = 24
        self.object._perform_year_end_procedure(self.clock)

        self.assertEqual(len(self.object.gl.transactions), 20)
        self.assertEqual(self.object.gl.transactions[14].amount, 80)
        self.assertEqual(self.object.gl.transactions[14].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[14].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[14].cr_account,
                         self.sales_acc.path)

        self.assertEqual(self.object.gl.transactions[15].amount, 20)
        self.assertEqual(self.object.gl.transactions[15].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[15].dt_account,
                         self.cos_acc.path)
        self.assertEqual(self.object.gl.transactions[15].cr_account,
                         gls._acci_gross_prof_.path)

        self.assertEqual(self.object.gl.transactions[16].amount, 800)
        self.assertEqual(self.object.gl.transactions[16].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[16].dt_account,
                         self.reva_acc.path)
        self.assertEqual(self.object.gl.transactions[16].cr_account,
                         gls._acci_inc_sum_.path)

        self.assertEqual(self.object.gl.transactions[17].amount, 500)
        self.assertEqual(self.object.gl.transactions[17].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[17].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[17].cr_account,
                         self.expa_acc.path)

        self.assertEqual(self.object.gl.transactions[18].amount, 60)
        self.assertEqual(self.object.gl.transactions[18].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[18].dt_account,
                         gls._acci_gross_prof_.path)
        self.assertEqual(self.object.gl.transactions[18].cr_account,
                         gls._acci_inc_sum_.path)

        self.assertEqual(self.object.gl.transactions[19].amount, 360)
        self.assertEqual(self.object.gl.transactions[19].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[19].dt_account,
                         gls._acci_inc_sum_.path)
        self.assertEqual(self.object.gl.transactions[19].cr_account,
                         gls._accb_ret_earnings_acc_.path)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module contains code used to test core object classes.
"""

import unittest

import jsonpickle

from auxi.core.objects import Object, NamedObject


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ObjectUnitTester(unittest.TestCase):
    """
    The unit tester for the class being tested.
    """

    def setUp(self):
        self.o = Object()

    def tearDown(self):
        del self.o

    def test___str__(self):
        """
        Test whether the __str__ method successfully generates a json string
        representation of the object.
        """

        # Confirm that the string representation of the current object and that
        # of an object decoded from this string representation are equal.
        str_o = str(self.o)
        new_o = jsonpickle.decode(str_o)
        self.assertEqual(str_o, str(new_o))


class NamedObjectUnitTester(unittest.TestCase):
    """
    The unit tester for the class being tested.
    """

    def setUp(self):
        self.o = NamedObject('NameA', 'DescriptionA')

    def tearDown(self):
        del self.o

    def test_constructor(self):
        """
        Test whether the constructor successfully initialises the object.
        """

        self.assertEqual(self.o.name, 'NameA')
        self.assertEqual(self.o.description, 'DescriptionA')

    def test__str__(self):
        """
        Test whether the __str__ method successfully generates a json string
        representation of the object.
        """

        # Confirm that the string representation of the current object and that
        # of an object decoded from this string representation are equal.
        str_o = str(self.o)
        new_o = jsonpickle.decode(str_o)
        self.assertEqual(str_o, str(new_o))

        # Confirm that the property values are preserved through string
        # encoding and decoding.
        self.assertEqual(new_o.name, 'NameA')
        self.assertEqual(new_o.description, 'DescriptionA')


if __name__ == '__main__':
    unittest.main()



"""
This package provides tools to do calculations related to heat transfer.
"""

from math import radians, cos
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle

from auxi.core.objects import Object, NamedObject
from auxi.tools.transportphenomena import dimensionlessquantities as dq
from auxi.tools import physicalconstants as const


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class EmpiricalCorrelation(Object):
    """
    Abstract base class for empirical correlations to calculate natural
    convection heat transfer.
    """

    def __init__(self, fluid, references):
        self._fluid = fluid
        self.references = references


class IsothermalFlatSurface(EmpiricalCorrelation):
    """
    Contains configuration and calculations for an isothermal flat surface.

    :param fluid: material physical properties of a fluid
    :param isgas: Specify if the surface is a gas (default is True)
    :param references: The references used for the configurations
      (default is None)
    """

    class Region(NamedObject):
        """
        Defines a region in the 2D Ra-theta coordinate system. A region is used
        to determine which equation should be used to determine the isothermal
        flat surface.

        :param name: name of the region
        :param theta_min: minimum theta value
        :param theta_min_incl: should the minimum theta be included?
        :param theta_max: maximum theta value
        :param theta_max_incl: should the maximum theta be included?
        :param Ra_min: minimum Ra value (default is None)
        :param Ra_min_incl: should the minimum Ra be included? (default is True)
        :param Ra_max: maximum Ra value (default is None)
        :param Ra_max_incl: should the maximum Ra be included? (default is True)
        :param description: description of the region (default is None)
        """

        def __init__(self, name,
                     theta_min, theta_min_incl,
                     theta_max, theta_max_incl,
                     Ra_min=None, Ra_min_incl=True,
                     Ra_max=None, Ra_max_incl=True,
                     description=None):
            super().__init__(name, description)

            self.theta_min = theta_min
            self.theta_min_incl = theta_min_incl
            if self.theta_min_incl:
                self._theta_min = self._theta_min_incl
            else:
                self._theta_min = self._theta_min_excl

            self.theta_max = theta_max
            self.theta_max_incl = theta_max_incl
            if self.theta_max_incl:
                self._theta_max = self._theta_max_incl
            else:
                self._theta_max = self._theta_max_excl

            self.Ra_min = Ra_min
            self.Ra_min_incl = Ra_min_incl
            if self.Ra_min is None:
                self._Ra_min = self._Ra_min_none
            elif self.Ra_min_incl:
                self._Ra_min = self._Ra_min_incl
            else:
                self._Ra_min = self._Ra_min_excl

            self.Ra_max = Ra_max
            self.Ra_max_incl = Ra_max_incl
            if self.Ra_max is None:
                self._Ra_max = self._Ra_max_none
            elif self.Ra_max_incl:
                self._Ra_max = self._Ra_max_incl
            else:
                self._Ra_max = self._Ra_max_excl

        def _theta_min_incl(self, theta):
            return theta >= self.theta_min

        def _theta_min_excl(self, theta):
            return theta > self.theta_min

        def _theta_max_incl(self, theta):
            return theta <= self.theta_max

        def _theta_max_excl(self, theta):
            return theta < self.theta_max

        def _Ra_min_none(self, Ra):
            return True

        def _Ra_min_incl(self, Ra):
            return Ra >= self.Ra_min

        def _Ra_min_excl(self, Ra):
            return Ra > self.Ra_min

        def _Ra_max_none(self, Ra):
            return True

        def _Ra_max_incl(self, Ra):
            return Ra <= self.Ra_max

        def _Ra_max_excl(self, Ra):
            return Ra < self.Ra_max

        def contains_point(self, theta, Ra):
            """
            Determines if a theta and Ra value falls inside the region.

            :param theta: theta value
            :param Ra: Ra value

            :returns: True of False
            """

            if not self._theta_min(theta):
                return False
            if not self._theta_max(theta):
                return False
            if not self._Ra_min(Ra):
                return False
            if not self._Ra_max(Ra):
                return False
            return True

        def plot_region(self):
            """
            Plots the region to a graph
            """

            theta_diff = self.theta_max - self.theta_min
            Ra_min_plot = self.Ra_min
            Ra_max_plot = self.Ra_max
            if self.Ra_min is None:
                Ra_min_plot = 3
                Ra_diff = Ra_max_plot - Ra_min_plot
            elif self.Ra_max is None:
                Ra_max_plot = 12
                Ra_diff = Ra_max_plot - Ra_min_plot
            else:
                Ra_diff = Ra_max_plot - Ra_min_plot

            plt.xlim(-90, 90)
            plt.ylim(3, 12)

            plt.xlabel('angle [°]')
            plt.ylabel('Ra')

            currentAxis = plt.gca()

            if self.description == "In_Eq_Region":
                currentAxis.add_patch(Rectangle((self.theta_min, Ra_min_plot),
                                                theta_diff, Ra_diff,
                                                facecolor="grey"))
            else:
                currentAxis.add_patch(Rectangle((self.theta_min, Ra_min_plot),
                                                theta_diff, Ra_diff,
                                                facecolor="red"))

            mid_theta = (self.theta_max + self.theta_min-5) / 2.0
            mid_Ra = (Ra_max_plot + Ra_min_plot-0.6) / 2.0
            currentAxis.annotate(self.name, xy=(mid_theta, mid_Ra),
                                 xytext=(mid_theta, mid_Ra), fontsize=14)

    def __init__(self, fluid, isgas=True, references=None):
        super().__init__(fluid, references)

        self._isgas = isgas

        self.equation_dict = {}

        ifs = IsothermalFlatSurface
        eq_dict = self.equation_dict

        region = self.Region("r1", -90, True, -60, True, None, False, 4, False)
        eq_dict[region] = ifs._Nu_x__9_22
        region = self.Region("r2", -90, True, -60, True, 4, True, 7, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__9_22
        region = self.Region("r3", -90, True, -60, True, 7, False, 11, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__9_23
        region = self.Region("r4", -90, True, -60, True, 11, False, None,
                             False)
        eq_dict[region] = ifs._Nu_x__9_23
        region = self.Region("r5", -60, False, -52, True, None, False, 7, True)
        eq_dict[region] = ifs._Nu_x__9_22
        region = self.Region("r6", -60, False, -52, True, 7, False, None,
                             False)
        eq_dict[region] = ifs._Nu_x__9_23
        region = self.Region("r7", -52, False, -45, False, None, False, 7,
                             True)
        eq_dict[region] = ifs._Nu_x__8_27_g
        region = self.Region("r8", -52, False, -45, False, 7, False, None,
                             False)
        eq_dict[region] = ifs._Nu_x__8_27_g
        region = self.Region("r9", -45, True, 0, False, None, False, 5, False)
        eq_dict[region] = ifs._Nu_x__8_27_g
        region = self.Region("r10", -45, True, 0, False, 5, True, 9, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__8_27_g
        region = self.Region("r11", -45, True, 0, False, 9, True, None, True)
        eq_dict[region] = ifs._Nu_x__8_27_g
        region = self.Region("r12", 0, True, 0, True, None, False, 5, True)
        eq_dict[region] = ifs._Nu_x__8_26
        region = self.Region("r13", 0, True, 0, True, 5, False, 11, False,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__8_26
        region = self.Region("r14", 0, True, 0, True, 11, False, None, False)
        eq_dict[region] = ifs._Nu_x__8_26
        region = self.Region("r15", 0, False, 88, False, None, False, 5, False)
        eq_dict[region] = ifs._Nu_x__8_27
        region = self.Region("r16", 0, True, 88, False, 5, True, 11, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__8_27
        region = self.Region("r17", 0, False, 88, False, 11, False, None,
                             False)
        eq_dict[region] = ifs._Nu_x__8_27
        region = self.Region("r18", 88, False, 90, True, None, False, 6, False)
        eq_dict[region] = ifs._Nu_x__8_38
        region = self.Region("r19", 88, True, 90, True, 6, True, 9, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__8_38
        region = self.Region("r20", 88, True, 89, False, 9, False, None, False)
        eq_dict[region] = ifs._Nu_x__8_38
        region = self.Region("r21", 89, True, 90, True, 9, True, 11, True,
                             "In_Eq_Region")
        eq_dict[region] = ifs._Nu_x__8_38
        region = self.Region("r22", 88, False, 90, True, 11, False, None,
                             False)
        eq_dict[region] = ifs._Nu_x__8_38

        self.regions = list(self.equation_dict.keys())

    def _Nu_x__8_26(self, Ra, Pr):
        return (0.678 * Ra**0.25) * (Pr / (0.952 + Pr)) ** 0.25

    def _Nu_x__8_27(self, Ra, Pr):
        return (0.68 + (0.67 * Ra**0.25) / ((1.0 + (0.492 / Pr) **
                (9.0/16.0)) ** (4.0/9.0)))

    def _Nu_x__8_27_g(self, Ra, Pr):
        return (0.68 + (0.67 * Ra**0.25) / ((1.0 + (0.492 / Pr) **
                (9.0/16.0)) ** (4.0/9.0)))

    def _Nu_x__8_38(self, Ra, Pr):
        return 0.58 * Ra**0.2

    def _Nu_x__9_22(self, Ra, Pr):
        return 0.59 * Ra**0.25

    def _Nu_x__9_23(self, Ra, Pr):
        return 0.1 * Ra**(1/3.0)

    def Nu_x(self, L, theta, Ts, **statef):
        """
        Calculate the local Nusselt number.

        :param L: [m] characteristic length of the heat transfer surface
        :param theta: [°] angle of the surface with the vertical
        :param Ts: [K] heat transfer surface temperature
        :param Tf: [K] bulk fluid temperature

        :returns: float
        """

        Tf = statef['T']
        thetar = radians(theta)

        if self._isgas:
            self.Tr = Ts - 0.38 * (Ts - Tf)
            beta = self._fluid.beta(T=Tf)
        else:  # for liquids
            self.Tr = Ts - 0.5 * (Ts - Tf)
            beta = self._fluid.beta(T=self.Tr)

        if Ts > Tf:  # hot surface
            if 0.0 < theta < 45.0:
                g = const.g*cos(thetar)
            else:
                g = const.g
        else:  # cold surface
            if -45.0 < theta < 0.0:
                g = const.g*cos(thetar)
            else:
                g = const.g

        nu = self._fluid.nu(T=self.Tr)
        alpha = self._fluid.alpha(T=self.Tr)

        Gr = dq.Gr(L, Ts, Tf, beta, nu, g)
        Pr = dq.Pr(nu, alpha)
        Ra = Gr * Pr

        eq = [self.equation_dict[r]
              for r in self.regions if r.contains_point(theta, Ra)][0]

        return eq(self, Ra, Pr)

    def Nu_L(self, L, theta, Ts, **statef):
        """
        Calculate the average Nusselt number.

        :param L: [m] characteristic length of the heat transfer surface
        :param theta: [°] angle of the surface with the vertical
        :param Ts: [K] heat transfer surface temperature
        :param **statef: [K] bulk fluid temperature

        :returns: float
        """

        return self.Nu_x(L, theta, Ts, **statef) / 0.75

    def h_x(self, L, theta, Ts, **statef):
        """
        Calculate the local heat transfer coefficient.

        :param L: [m] characteristic length of the heat transfer surface
        :param theta: [°] angle of the surface with the vertical
        :param Ts: [K] heat transfer surface temperature
        :param Tf: [K] bulk fluid temperature

        :returns: [W/m2/K] float
        """

        Nu_x = self.Nu_x(L, theta, Ts, **statef)
        k = self._fluid.k(T=self.Tr)
        return Nu_x * k / L

    def h_L(self, L, theta, Ts, **statef):
        """
        Calculate the average heat transfer coefficient.

        :param L: [m] characteristic length of the heat transfer surface
        :param theta: [°] angle of the surface with the vertical
        :param Ts: [K] heat transfer surface temperature
        :param Tf: [K] bulk fluid temperature

        :returns: [W/m2/K] float
        """

        Nu_L = self.Nu_L(L, theta, Ts, **statef)
        k = self._fluid.k(T=self.Tr)
        return Nu_L * k / L

    def plot_regions(self):
        """
        Plot all the regions.
        """
        for region in self.regions:
            region.plot_region()
        plt.show()



#!/usr/bin/env python3
"""
This module contains code used to test core reporting classes.
"""

import unittest
import os.path

from auxi.core.reporting import ReportFormat, Report


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ReportingUnitTester(unittest.TestCase):
    """
    The unit tester for the class being tested.
    """

    def setUp(self):
        self.object = Report([["cola", "colb"], [1, 2]])
        self.object_w_file = Report([["cola", "colb"], [1, 2]], "./test")

    def tearDown(self):
        del self.object

    def test_render_without_file(self):
        """
        Test whether a value is returned when a report is rendered in a
        string format. And wether any of the other formats throw an error.
        """

        self.object.render(format=ReportFormat.printout)
        self.object.render(format=ReportFormat.latex)
        self.object.render(format=ReportFormat.txt)
        self.object.render(format=ReportFormat.csv)

        self.assertEqual(
            len(self.object.render(format=ReportFormat.string)) > 0,
            True)

        self.object.render(format=ReportFormat.matplotlib)
        self.object.render(format=ReportFormat.png)

    def test_render_with_file(self):
        """
        Test whether a value is returned when a report is rendered in a
        string format. And wether any of the other formats throw an error.
        """

        file_name = self.object_w_file.output_path
        self.object_w_file.render(format=ReportFormat.printout)
        self.object_w_file.render(format=ReportFormat.latex)
        os.path.isfile(file_name + '.tex')
        os.remove(file_name + '.tex')
        self.object_w_file.render(format=ReportFormat.txt)
        os.path.isfile(file_name + '.txt')
        os.remove(file_name + '.txt')
        self.object_w_file.render(format=ReportFormat.csv)
        os.path.isfile(file_name + '.csv')
        os.remove(file_name + '.csv')

        self.assertEqual(
            len(self.object_w_file.render(format=ReportFormat.string)) > 0,
            True)

        self.object_w_file.render(format=ReportFormat.matplotlib)
        self.object_w_file.render(format=ReportFormat.png)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module contains code used to test core helpers classes.
"""

import unittest
import os
from datetime import datetime, date

from auxi.core.helpers import get_path_relative_to_module, get_date

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class HelpersUnitTester(unittest.TestCase):
    """
    The unit tester for the helpers functions being tested.
    """

    def test_get_path_relative_to_module(self):
        path = get_path_relative_to_module(__file__, "test/test.txt")
        dr = os.path.dirname(__file__)
        self.assertEqual(
            path,
            os.path.join(dr, "test/test.txt"))

    def test_get_date(self):
        dt = date(2016, 4, 14)
        self.assertEqual(get_date(dt), dt)
        self.assertEqual(get_date("2016-04-14"), dt)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes and functions for doing thermochemical
calculations.
"""

import os
import sys
import glob
import math

from auxi.core.objects import Object, NamedObject
from auxi.core.helpers import get_path_relative_to_module as get_path
from auxi.tools.chemistry.stoichiometry import molar_mass as mm


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class CpRecord(Object):
    """
    A heat capacity (Cp) equation record for a compound phase over a
    specific temperature range.

    :param dictionary: A dictionary containing the data required to initialise
      the phase.
    """

    def __init__(self, dictionary):
        self.Tmin = dictionary['Tmin']
        """[K] The minimum temperature of the range covered by this record."""
        if type(self.Tmin) is not float:
            self.Tmin = float(self.Tmin)

        self.Tmax = dictionary['Tmax']
        """[K] The maximum temperature of the range covered by this record."""
        if type(self.Tmax) is not float:
            self.Tmax = float(self.Tmax)

        self._coefficients = []
        """The coefficiencts of the terms in the Cp equation."""

        self._exponents = []
        """The exponents of the terms in the Cp equation."""

        for t in dictionary['Terms']:
            c = t['Coefficient']
            e = t['Exponent']
            if type(c) is not float:
                c = float(c)
            if type(e) is not float:
                e = float(e)
            self._coefficients.append(c)
            self._exponents.append(e)

    def __str__(self):
        result = '\t\tCp RECORD:' + '\n'
        result += '\t\t\tTmin: ' + str(self.Tmin) + '\n'
        result += '\t\t\tTmax: ' + str(self.Tmax) + '\n'

        for i in range(len(self._coefficients)):
            result += '\t\t\t{:.8e}'.format(self._coefficients[i]) + ' '
            result += str(self._exponents[i]) + '\n'

        return result

    def Cp(self, temperature):
        """
        Calculate the heat capacity of the compound phase.

        :param temperature: [K]

        :returns: [J/mol/K] Heat capacity.
        """

        result = 0.0
        for c, e in zip(self._coefficients, self._exponents):
            result += c * temperature ** e
        return result

    def H(self, temperature):
        """
        Calculate the portion of enthalpy of the compound phase covered by this
        Cp record.

        :param temperature: [K]

        :returns: [J/mol] Enthalpy.
        """

        result = 0.0
        if temperature < self.Tmax:
            T = temperature
        else:
            T = self.Tmax
        Tref = self.Tmin
        for c, e in zip(self._coefficients, self._exponents):
            # Analytically integrate Cp(T).
            if e == -1.0:
                result += c * (math.log(T) - math.log(Tref))
            else:
                result += c * (T ** (e + 1.0) - Tref ** (e + 1.0)) / (e + 1.0)
        return result

    def S(self, temperature):
        """
        Calculate the portion of entropy of the compound phase covered by this
        Cp record.

        :param temperature: [K]

        :returns: Entropy. [J/mol/K]
        """

        result = 0.0
        if temperature < self.Tmax:
            T = temperature
        else:
            T = self.Tmax
        Tref = self.Tmin
        for c, e in zip(self._coefficients, self._exponents):
            # Create a modified exponent to analytically integrate Cp(T)/T
            # instead of Cp(T).
            e_modified = e - 1.0
            if e_modified == -1.0:
                result += c * (math.log(T) - math.log(Tref))
            else:
                e_mod = e_modified + 1.0
                result += c * (T ** (e_mod) - Tref ** (e_mod))  \
                    / (e_mod)
        return result


class Phase(NamedObject):
    """
    A phase of a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise
      the phase.
    """

    def __init__(self, dictionary):
        self.name = dictionary['Symbol']
        """The phase's name, e.g. solid, liquid, gas, etc."""

        self.symbol = dictionary['Symbol']
        """The phase's symbol, e.g. S1 = solid 1, L = liquid, etc."""

        self.Tref = 298.15
        """[K] The reference temperature of the phase."""

        self.DHref = dictionary['DHref']
        """[J/mol] The formation enthalpy of the phase at Tref."""

        self.Sref = dictionary['Sref']
        """[J/mol/K] The standard entropy of the phase at Tref."""

        self._Cp_records = {}
        """A dictionary containing the phase's Cp records."""

        for k, v in dictionary['Cp_records'].items():
            tmax = k
            if type(tmax) is not float:
                tmax = float(tmax)
            self._Cp_records[tmax] = CpRecord(v)

    def __str__(self):
        result = '\tPHASE: ' + self.name + '\n'
        result += '\t\tName: ' + self.name + '\n'
        result += '\t\tSymbol: ' + self.symbol + '\n'
        result += '\t\tTref: ' + str(self.Tref) + '\n'
        result += '\t\tDHref: ' + str(self.DHref) + '\n'
        result += '\t\tSref: ' + str(self.Sref) + '\n'
        result += '\t\tCp record count:' + str(len(self._Cp_records)) + '\n'

        for k, v in self._Cp_records.items():
            result += str(v)

        return result

    def Cp(self, temperature):
        """
        Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: [J/mol/K] The heat capacity of the compound phase.
        """

        for Tmax in sorted(self._Cp_records.keys()):
            if temperature < Tmax:
                return self._Cp_records[Tmax].Cp(temperature)

        Tmax = max(self._Cp_records.keys())

        return self._Cp_records[Tmax].Cp(Tmax)

    def H(self, temperature):
        """
        Calculate the enthalpy of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: [J/mol] The enthalpy of the compound phase.
        """

        result = self.DHref

        for Tmax in sorted(self._Cp_records.keys()):
            result += self._Cp_records[Tmax].H(temperature)
            if temperature <= Tmax:
                return result

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        result += self.Cp(Tmax)*(temperature - Tmax)

        return result

    def S(self, temperature):
        """
        Calculate the entropy of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: [J/mol/K] The entropy of the compound phase.
        """

        result = self.Sref

        for Tmax in sorted(self._Cp_records.keys()):
            result += self._Cp_records[Tmax].S(temperature)
            if temperature <= Tmax:
                return result

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        result += self.Cp(Tmax)*math.log(temperature / Tmax)

        return result

    def G(self, temperature):
        """Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: [J/mol] The Gibbs free energy of the compound phase.
        """

        h = self.DHref
        s = self.Sref

        for Tmax in sorted(self._Cp_records.keys()):
            h = h + self._Cp_records[Tmax].H(temperature)
            s = s + self._Cp_records[Tmax].S(temperature)
            if temperature <= Tmax:
                return h - temperature * s

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        h = h + self.Cp(Tmax)*(temperature - Tmax)
        s = s + self.Cp(Tmax)*math.log(temperature / Tmax)

        return h - temperature * s


class Compound(Object):
    """
    Represents a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise
      the compound.
    """

    def __init__(self, dictionary):
        self.formula = dictionary['Formula']
        """Chemical formula, e.g. 'Fe', 'CO2'."""

        self.molar_mass = mm(self.formula) / 1000.0
        """Molar mass. [kg/mol]"""

        self._phases = {}
        """Dictionary containing the compound's phase objects."""

        if 'Reference' in dictionary:
            self.reference = dictionary['Reference']
        else:
            self.reference = ""

        """Reference to the publisher of the thermo data."""

        for k, v in dictionary['Phases'].items():
            self._phases[k] = Phase(v)

#    def __str__(self):
#        result = 'COMPOUND: ' + '\n'
#        result += '\tFormula: ' + self.formula + '\n'
#
#        for k, v in self._phases.items():
#            result += str(v)
#
#        return result

    def get_phase_list(self):
        """
        Get a list of the compound's phases.

        :returns: List of phases.
        """

        return sorted(self._phases.keys())

    def get_reference(self):
        if self.reference in [None, ""]:
            return ""
        return get_datafile_references()[self.reference]

    def Cp(self, phase, temperature):
        """
        Calculate the heat capacity of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param temperature: [K]

        :returns: [J/mol/K] Heat capacity.
        """

        try:
            return self._phases[phase].Cp(temperature)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))

    def H(self, phase, temperature):
        """
        Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param temperature: [K]

        :returns: [J/mol] Enthalpy.
        """

        try:
            return self._phases[phase].H(temperature)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))

    def S(self, phase, temperature):
        """
        Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param temperature: [K]

        :returns: [J/mol/K] Entropy.
        """

        try:
            return self._phases[phase].S(temperature)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))

    def G(self, phase, temperature):
        """
        Calculate the Gibbs free energy of a phase of the compound at a
        specified temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param temperature: [K]

        :returns: [J/mol] Gibbs free energy.
        """

        try:
            return self._phases[phase].G(temperature)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))


def _get_default_data_path_():
    """
    Calculate the default path in which thermochemical data is stored.

    :returns: Default path.
    """

    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r'data/rao')
    data_path = os.path.abspath(data_path)
    return data_path


def _read_compound_from_factsage_file_(file_name):
    """
    Build a dictionary containing the factsage thermochemical data of a
    compound by reading the data from a file.

    :param file_name: Name of file to read the data from.

    :returns: Dictionary containing compound data.
    """

    with open(file_name) as f:
        lines = f.readlines()

    compound = {}
    compound['Formula'] = lines[0].split(' ')[1]
    compound['Phases'] = {}
    phs = compound['Phases']

    started = False
    phaseold = 'zz'
    recordold = '0'

    for line in lines:
        if started:
            if line.startswith('_'):  # line indicating end of data
                break
            line = line.replace(' 298 ', ' 298.15 ')
            line = line.replace(' - ', ' ')
            while '  ' in line:
                line = line.replace('  ', ' ')
            line = line.replace(' \n', '')
            line = line.replace('\n', '')
            strings = line.split(' ')
            if len(strings) < 2:  # empty line
                continue
            phase = strings[0]
            if phase != phaseold:  # new phase detected
                phaseold = phase
                ph = phs[phase] = {}
                ph['Symbol'] = phase
                ph['DHref'] = float(strings[2])
                ph['Sref'] = float(strings[3])
                cprecs = ph['Cp_records'] = {}
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    cprecs[Tmax] = {}
                    cprecs[Tmax]['Tmin'] = float(strings[len(strings) - 2])
                    cprecs[Tmax]['Tmax'] = float(strings[len(strings) - 1])
                    cprecs[Tmax]['Terms'] = []
                    t = {}
                    t['Coefficient'] = float(strings[4])
                    t['Exponent'] = float(strings[5])
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 10:
                        t = {}
                        t['Coefficient'] = float(strings[6])
                        t['Exponent'] = float(strings[7])
                        cprecs[Tmax]['Terms'].append(t)
                else:  # old record detected
                    t = {}
                    t['Coefficient'] = float(strings[2])
                    t['Exponent'] = float(strings[3])
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {}
                        t['Coefficient'] = float(strings[4])
                        t['Exponent'] = float(strings[5])
                        cprecs[Tmax]['Terms'].append(t)
            else:  # old phase detected
                ph = phs[phase]
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    cprecs = ph['Cp_records']
                    cprecs[Tmax] = {}
                    cprecs[Tmax]['Tmin'] = float(strings[len(strings) - 2])
                    cprecs[Tmax]['Tmax'] = float(strings[len(strings) - 1])
                    cprecs[Tmax]['Terms'] = []
                    t = {}
                    t['Coefficient'] = float(strings[2])
                    t['Exponent'] = float(strings[3])
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {}
                        t['Coefficient'] = float(strings[4])
                        t['Exponent'] = float(strings[5])
                        cprecs[Tmax]['Terms'].append(t)
                else:  # old record detected
                    t = {}
                    t['Coefficient'] = float(strings[2])
                    t['Exponent'] = float(strings[3])
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {}
                        t['Coefficient'] = float(strings[4])
                        t['Exponent'] = float(strings[5])
                        cprecs[Tmax]['Terms'].append(t)
        if line.startswith('_'):  # line indicating the start of the data
            started = True

    return compound


def _split_compound_string_(compound_string):
    """
    Split a compound's combined formula and phase into separate strings for
    the formula and phase.

    :param compound_string: Formula and phase of a chemical compound, e.g.
      'SiO2[S1]'.

    :returns: Formula of chemical compound.
    :returns: Phase of chemical compound.
    """

    formula = compound_string.replace(']', '').split('[')[0]
    phase = compound_string.replace(']', '').split('[')[1]

    return formula, phase


def _finalise_result_(compound, value, mass):
    """
    Convert the value to its final form by unit conversions and multiplying
    by mass.

    :param compound: Compound object.
    :param value: [J/mol] Value to be finalised.
    :param mass: [kg] Mass of compound.

    :returns: [kWh] Finalised value.
    """

    result = value / 3.6E6  # J/x -> kWh/x
    result = result / compound.molar_mass  # x/mol -> x/kg
    result = result * mass  # x/kg -> x

    return result


def _read_compound_from_auxi_file_(file_name):
    """
    Build a dictionary containing the auxi thermochemical data of a compound by
    reading the data from a file.

    :param file_name: Name of file to read the data from.

    :returns: Dictionary containing compound data.
    """

    with open(file_name) as f:
        content = eval(f.read())
    return content


def write_compound_to_auxi_file(directory, compound):
    """
    Writes a compound to an auxi file at the specified directory.

    :param dir: The directory.
    :param compound: The compound.
    """

    file_name = "Compound_" + compound.formula + ".json"
    with open(os.path.join(directory, file_name), 'w') as f:
        f.write(str(compound))


def load_data_factsage(path=''):
    """
    Load all the thermochemical data factsage files located at a path.

    :param path: Path at which the data files are located.
    """

    compounds.clear()

    if path == '':
        path = default_data_path
    if not os.path.exists(path):
        raise Exception('The path does not exist.')

    files = glob.glob(os.path.join(path, 'Compound_*.txt'))

    for file in files:
        compound = Compound(_read_compound_from_factsage_file_(file))
        compounds[compound.formula] = compound


def load_data_auxi(path=''):
    """
    Load all the thermochemical data auxi files located at a path.

    :param path: Path at which the data files are located.
    """

    compounds.clear()

    if path == '':
        path = default_data_path
    if not os.path.exists(path):
        raise Exception('The path does not exist.')

    files = glob.glob(os.path.join(path, 'Compound_*.json'))

    for file in files:
        compound = Compound(_read_compound_from_auxi_file_(file))
        compounds[compound.formula] = compound


def list_compounds():
    """
    List all compounds that are currently loaded in the thermo module, and
    their phases.
    """

    print('Compounds currently loaded:')
    for compound in sorted(compounds.keys()):
        phases = compounds[compound].get_phase_list()
        for phase in phases:
            print(compound + '[' + phase + ']')


def get_datafile_references():
    """
    Retrieve all the references used by the datafiles.
    """

    with open(get_path(__file__, "data/references.json")) as f:
        content = eval(f.read())

    return content


def molar_mass(compound):
    """
    Determine the molar mass of a chemical compound.

    :param compound: Formula of a chemical compound, e.g. 'Fe2O3'.

    :returns: [kg/mol] Molar mass.
    """

    return mm(compound) / 1000.0


def Cp(compound_string, temperature, mass=1.0):
    """
    Calculate the heat capacity of the compound for the specified temperature
    and mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param temperature: [°C]
    :param mass: [kg]

    :returns: [kWh/K] Heat capacity.
    """

    formula, phase = _split_compound_string_(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.Cp(phase, temperature_K)

    return _finalise_result_(compound, result, mass)


def H(compound_string, temperature, mass=1.0):
    """
    Calculate the enthalpy of the compound for the specified temperature and
    mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param temperature: [°C]
    :param mass: [kg]

    :returns: [kWh] Enthalpy.
    """

    formula, phase = _split_compound_string_(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.H(phase, temperature_K)

    return _finalise_result_(compound, result, mass)


def S(compound_string, temperature, mass=1.0):
    """
    Calculate the entropy of the compound for the specified temperature and
    mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param temperature: [°C]
    :param mass: [kg]

    :returns: [kWh/K] Entropy.
    """

    formula, phase = _split_compound_string_(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.S(phase, temperature_K)

    return _finalise_result_(compound, result, mass)


def G(compound_string, temperature, mass=1.0):
    """
    Calculate the Gibbs free energy of the compound for the specified
    temperature and mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param temperature: [°C]
    :param mass: [kg]


    :returns: [kWh] Gibbs free energy.
    """

    formula, phase = _split_compound_string_(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.G(phase, temperature_K)

    return _finalise_result_(compound, result, mass)


compounds = {}
default_data_path = _get_default_data_path_()
load_data_auxi()


if __name__ == '__main__':
    import unittest
    from auxi.tools.chemistry.thermochemistry_test import ThermoFunctionTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides tools for calculating physical properties of an ideal gas.
"""


from auxi.tools.materialphysicalproperties.core import Model
from auxi.tools.physicalconstants import R
from auxi.tools.chemistry.stoichiometry import molar_mass as mm


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Johan Zietsman'
__credits__ = ['Johan Zietsman']
__maintainer__ = 'Johan Zietsman'
__email__ = 'johan.zietsman@ex-mente.co.za'
__status__ = 'Planning'


class BetaT(Model):
    """
    A model that describes the variation in the thermal expansion coefficient
    of an ideal gas as a function of temperature.
    """

    def __init__(self):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0}}
        super().__init__('Ideal Gas', 'Thermal Expansion Coefficient', 'beta',
                         '\\beta', '1/K', state_schema, None, None)

    def calculate(self, **state):
        """
        Calculate the thermal expansion coefficient at the specified
        temperature:

        :param T: [K] temperature

        :returns: [1/K] thermal expansion coefficient

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """
        super().calculate(**state)
        return 1.0 / state['T']


class RhoT(Model):
    """
    A model that describes the variation in density of an ideal gas as a
    function of temperature.

    :param molar_mass: [g/mol] average molar mass of the gas
    :param P: [Pa] pressure
    """

    def __init__(self, molar_mass, P):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0}}
        super().__init__('Ideal Gas', 'Density', 'rho', '\\rho', 'kg/m3',
                         state_schema, None, None)

        self.mm = molar_mass / 1000.0
        """[kg/mol] average molar mass of the gas"""
        self.P = P
        """[Pa] pressure"""

    def calculate(self, **state):
        """
        Calculate the density at the specified temperature.

        :param T: [K] temperature

        :returns: [kg/m3] density

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """
        super().calculate(**state)
        return self.mm * self.P / R / state['T']


class RhoTP(Model):
    """
    A model that describes the variation in density of an ideal gas as a
    function of temperature and pressure.

    :param molar_mass: [g/mol] average molar mass of the gas
    """

    def __init__(self, molar_mass):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'P': {'required': True, 'type': 'float', 'min': 0.0}}
        super().__init__('Ideal Gas', 'Density', 'rho', '\\rho', 'kg/m3',
                         state_schema, None, None)

        self.mm = molar_mass / 1000.0
        """[kg/mol] average molar mass of the gas"""

    def calculate(self, **state):
        """
        Calculate the density at the specified temperature and pressure.

        :param T: [K] temperature
        :param P: [Pa] pressure

        :returns: [kg/m3] density

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """
        super().calculate(**state)
        return self.mm * state['P'] / R / state['T']


class RhoTPx(Model):
    """
    A model that describes the variation in density of an ideal gas as a
    function of temperature, pressure, and molar composition.
    """

    def __init__(self):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'P': {'required': True, 'type': 'float', 'min': 0.0},
                        'x': {'required': True, 'type': 'dict'}}
        super().__init__('Ideal Gas', 'Density', 'rho', '\\rho', 'kg/m3',
                         state_schema, None, None)

    def calculate(self, **state):
        """
        Calculate the density at the specified temperature, pressure, and
        composition.

        :param T: [K] temperature
        :param P: [Pa] pressure
        :param x: [mole fraction] dictionary of compounds and mole fractions

        :returns: [kg/m3] density

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """
        super().calculate(**state)
        mm_average = 0.0
        for compound, molefraction in state['x'].items():
            mm_average += molefraction * mm(compound)
        mm_average /= 1000.0

        return mm_average * state['P'] / R / state['T']


if __name__ == '__main__':
    import unittest
    from idealgas_test import *
    unittest.main()



from datetime import datetime
from dateutil.relativedelta import relativedelta

from auxi.core.time import TimePeriod
from auxi.modelling.business.models import TimeBasedModel
from auxi.modelling.business.basic import BasicActivity, BasicLoanActivity
from auxi.modelling.financial.des import GeneralLedgerStructure
from auxi.core.reporting import ReportFormat

# Create general ledger structure and accounts.
gl_structure = GeneralLedgerStructure("Courier GL Structure")

gl_structure["Long Term Borrowing"].create_account("Loan", "0000")
gl_structure["Expense"].create_account("Interest Expense", "0000")
gl_structure["Fixed Assets"].create_account("Vehicle Asset", "0000")
gl_structure["Sales"].create_account("Sales Delivery", "0000")
gl_structure["Cost of Sales"].create_account("Fuel", "0000")
gl_structure["Expense"].create_account("Wages", "0000")

# Create the business model, entity and components.
start_datetime = datetime(2016, 2, 1)
end_datetime = datetime(2021, 1, 1)

model = TimeBasedModel("Business Model", start_datetime=start_datetime,
                       period_duration=TimePeriod.month, period_count=61)

courier_company = model.create_entity("CourierZA", gl_structure=gl_structure)
ops = courier_company.create_component("Operations")
hr = courier_company.create_component("HR")

# Create activities
loan = BasicLoanActivity("Loan",
    bank_account="Bank/Default", loan_account="Long Term Borrowing/Loan",
    interest_account="Expense/Interest Expense",
    amount=200000, interest_rate=0.15, start=start_datetime, duration=36,
    interval=1)
ops.add_activity(loan)

purchase_vehicle = BasicActivity("Purchase Vehicle",
    dt_account="Fixed Assets/Vehicle Asset", cr_account="Bank/Default",
    amount=177000, start=start_datetime, end=start_datetime + relativedelta(months=1),
    interval=1)
ops.add_activity(purchase_vehicle)

make_delivery_sale = BasicActivity("Make Delivery",
    dt_account="Bank/Default", cr_account="Sales/Sales Delivery",
    amount=15000, start=start_datetime, end=end_datetime, interval=1)
ops.add_activity(make_delivery_sale)

pay_delivery_costs = BasicActivity("Pay for Fuel",
    dt_account="Cost of Sales/Fuel", cr_account="Bank/Default",
    amount=1000, start=start_datetime, end=end_datetime, interval=1)
ops.add_activity(pay_delivery_costs)

pay_wages = BasicActivity("Pay Wages", dt_account="Expense/Wages", cr_account="Bank/Default",
    amount=10000, start=start_datetime, end=end_datetime, interval=1)
hr.add_activity(pay_wages)

# Run the model
model.run()

# Print the reports.
#courier_company.gl.balance_sheet(format=ReportFormat.latex, output_path="balance_sheet.tex")
#courier_company.gl.income_statement(format=ReportFormat.latex, output_path="income_statement.tex")
courier_company.gl.transaction_list(component_path=ops.path)
courier_company.gl.balance_sheet()
courier_company.gl.income_statement(component_path=hr.path)



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
from os import remove
from os.path import realpath, dirname, join, isfile

from matplotlib import pyplot as plt

from auxi.tools.transportphenomena.heattransfer import naturalconvection as testee
from auxi.tools.materialphysicalproperties.gases import air


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


MODULE_PATH = dirname(realpath(__file__))


class DimensionlessQiantitiesTester(unittest.TestCase):

    def test_Gr(self):
        pass

    def test_Pr(self):
        pass

    def test_Re(self):
        pass

    def test_Ra(self):
        pass

    def test_Nu(self):
        pass

    def test_Sh(self):
        pass



#!/usr/bin/env python3
"""
This module runs all the tests of the auxi package at once.
"""

import unittest

from auxi.core.objects_test import ObjectUnitTester
from auxi.core.objects_test import NamedObjectUnitTester
from auxi.core.time_test import ClockUnitTester

from auxi.tools.chemistry.stoichiometry_test import StoichFunctionTester
from auxi.tools.chemistry.thermochemistry_test import ThermoFunctionTester
from auxi.tools.materialphysicalproperties.core_test import DataSetTester
from auxi.tools.materialphysicalproperties.idealgas_test \
    import BetaTTester, RhoTTester, RhoTPTester, RhoTPxTester
from auxi.tools.materialphysicalproperties.polynomial_test \
    import PolynomialModelTTester
from auxi.tools.transportphenomena.heattransfer.naturalconvection_test \
    import IsothermalFlatSurface_RegionTester, IsothermalFlatSurfaceTester
from auxi.tools.transportphenomena.dimensionlessquantities_test \
    import DimensionlessQiantitiesTester

from auxi.modelling.process.materials.chem_test \
    import ChemMaterialUnitTester, ChemMaterialPackageUnitTester
from auxi.modelling.process.materials.thermo_test \
    import ThermoMaterialUnitTester
# from auxi.modelling.process.materials.thermo_test \
#   import ThermoMaterialPackageUnitTester
from auxi.modelling.process.materials.psd_test \
    import PsdMaterialUnitTester, PsdMaterialPackageUnitTester
from auxi.modelling.process.materials.slurry_test \
    import SlurryMaterialUnitTester, SlurryMaterialPackageUnitTester


# MODELLING.FINANCIAL

from auxi.modelling.financial.des_test import GeneralLedgerAccountUnitTester
from auxi.modelling.financial.des_test import TransactionUnitTester
from auxi.modelling.financial.des_test import TransactionTemplateUnitTester
from auxi.modelling.financial.des_test import GeneralLedgerStructureUnitTester
from auxi.modelling.financial.des_test import GeneralLedgerUnitTester

from auxi.modelling.financial.reporting_test import GeneralLedgerStructureUnitTester
from auxi.modelling.financial.reporting_test import TransactionListUnitTester


# MODELLING.BUSINESS

from auxi.modelling.business.structure_test import ActivityUnitTester
from auxi.modelling.business.structure_test import ComponentUnitTester
from auxi.modelling.business.structure_test import EntityUnitTester

from auxi.modelling.business.basic_test import BasicActivityUnitTester
from auxi.modelling.business.basic_test import BasicLoanActivityUnitTester

from auxi.modelling.business.models_test import TimeBasedModelUnitTester


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes to work with materials and material packages that
are described with chemical compositions.
"""

import copy
from os.path import isfile

from auxi.core.objects import Object, NamedObject
from auxi.tools.chemistry.stoichiometry import element_mass_fractions as emf
from auxi.tools.chemistry import stoichiometry as stoich


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Material(NamedObject):
    """
    A material consisting of multiple chemical compounds.

    :param name: The material's name.
    :param file_path: The path of the material definition file.
    :param description: the material's description

    The format of the text file is as follows:

    * The lines are space separated. The values in a line are separated by \
      one or more spaces.
    * The first line is a heading line.
    * All subsequent lines contain a compound formula, followed by mass \
        fractions.
    * The first column lists the compounds in the material.
    * All subsequent columns describe assays of the material.

    The following is an example of a material text file::

        Compound   IlmeniteA  IlmeniteB  IlmeniteC
        Al2O3      0.01160    0.01550    0.00941
        CaO        0.00022    0.00001    0.00017
        Cr2O3      0.00008    0.00022    0.00011
        Fe2O3      0.20200    0.47300    0.49674
        Fe3O4      0.00000    0.00000    0.00000
        FeO        0.27900    0.19100    0.00000
        K2O        0.00004    0.00001    0.00005
        MgO        0.01040    0.00580    0.01090
        MnO        0.00540    0.00480    0.00525
        Na2O       0.00007    0.00005    0.00031
        P4O10      0.00001    0.00032    0.00015
        SiO2       0.00850    0.00490    0.01744
        TiO2       0.47700    0.29400    0.45949
        V2O5       0.00360    0.00800    0.00000
    """

    def __init__(self, name, file_path, description=None):
        self._validate_params_(name, file_path, description)
        super(Material, self).__init__(name, description)
        self._read_configuration_(file_path)

    def __str__(self):
        # TODO: Replace implementation with tabular.
        result = "Material: name='" + self.name + "'\n"

        # Create the header line of the table.
        result += 'Compound'.ljust(20)
        assay_names = sorted(self.assays.keys())
        for assay_name in assay_names:
            result += assay_name.ljust(20)
        result += '\n'

        # Create the content lines of the table.
        for compound in self.compounds:
            result += compound.ljust(20)
            compound_ix = self.get_compound_index(compound)
            for assay_name in assay_names:
                result += str(self.assays[assay_name][compound_ix]).ljust(20)
            result += '\n'
        return result

    def _validate_params_(self, name, file_path, description):
        super(Material, self)._validate_params_(name, description)

        if not isfile(file_path):
            raise ValueError('The specified file ({}) does not exist.'
                             .format(file_path))

    def _read_configuration_(self, file_path):
        self.compounds = []

        # Read the material's data from the file and prepare it for use.
        with open(file_path) as f:
            lines = f.readlines()
        lines = self._prepare_lines_(lines)

        # Determine assay names, and create a dictionary entry per assay.
        assay_names = lines[0].split(' ')
        del(assay_names[0:1])
        self.assays = dict()
        for assay_name in assay_names:
            self.assays[assay_name] = []

        # Read the compounds and assays.
        for i in range(1, len(lines)):
            strings = lines[i].split(' ')
            if len(strings) < len(assay_names) + 1:  # Not a full line.
                continue
            self.compounds.append(strings[0])  # Add the new compound.
            for j in range(0, len(self.assays)):  # Add mass fractions.
                assay_name = assay_names[j]
                self.assays[assay_name].append(float(strings[j+1]))
        self.compound_count = len(self.compounds)

        # Determine the list of elements.
        self.elements = self._create_element_list_()

    def _prepare_lines_(self, lines):
        """
        Prepare the lines read from the text file before starting to process
        it.
        """

        result = []
        for line in lines:
            # Remove all whitespace from the start and end of the line.
            line = line.strip()

            # Replace all tabs with spaces.
            line = line.replace('\t', ' ')

            # Replace all repeating spaces with a single space.
            while line.find('  ') > -1:
                line = line.replace('  ', ' ')

            result.append(line)

        return result

    def _create_element_list_(self):
        """
        Extract an alphabetically sorted list of elements from the compounds of
        the material.

        :returns: An alphabeticall sorted list of elements.
        """

        element_set = stoich.elements(self.compounds)
        return sorted(list(element_set))

    def get_compound_index(self, compound):
        """
        Determine the index of the specified compound.

        :param compound: The formula and phase of the specified compound, e.g.
          'Fe2O3[S1]'.

        :returns: The index of the specified compound.
        """

        return self.compounds.index(compound)

    def create_empty_assay(self):
        """
        Create an empty array to store an assay. The array's length will be
        equal to the number of compounds in the material.

        :returns: A floating point array.
        """

        return [0] * self.compound_count

    def add_assay(self, name, assay):
        """
        Add an assay to the material.

        :param name: The name of the new assay.
        :param assay: A list containing the compound mass fractions for
          the assay. The sequence of the assay's elements must correspond to
          the sequence of the material's compounds.
        """

        if not type(assay) is list:
            raise Exception('Invalid assay. It must be a list.')

        elif not len(assay) == self.compound_count:
            raise Exception('Invalid assay: It must have the same number of '
                            'elements as the material has compounds.')

        elif name in self.assays:
            raise Exception('Invalid assay: An assay with that name already '
                            'exists.')

        self.assays[name] = assay

    def get_assay_total(self, name):
        """
        Calculate the total of the specified assay.

        :param name: The name of the assay.

        :returns: The total mass fraction of the specified assay.
        """

        return sum(self.assays[name])

    def create_package(self, assay=None, mass=0.0, normalise=True):
        """
        Create a MaterialPackage based on the specified parameters.

        :param assay: The name of the assay based on which the package must be
          created.
        :param mass: [kg] The mass of the package.
        :param normalise: Indicates whether the assay must be normalised before
          creating the package.

        :returns: The created MaterialPackage.
        """

        if assay is None:
            return MaterialPackage(self, self.create_empty_assay())

        if normalise:
            assay_total = self.get_assay_total(assay)
        else:
            assay_total = 1.0
        masses = [(mass * m) / assay_total for m in self.assays[assay]]
        return MaterialPackage(self, masses)


class MaterialPackage(Object):
    """
    A package of a material consisting of multiple chemical compounds.

    :param material: A reference to the Material to which self belongs.
    :param compound_masses: [kg] The masses of the compounds in the package.
    """

    def __init__(self, material, compound_masses):
        self._validate_params_(material, compound_masses)

        self.material = material
        self.compound_masses = compound_masses

    def __str__(self):
        result = 'MaterialPackage\n'
        result += 'material'.ljust(20) + self.material.name + '\n'
        result += 'mass'.ljust(20) + str(self.get_mass()) + '\n'
        result += 'Component masses:\n'
        for compound in self.material.compounds:
            index = self.material.get_compound_index(compound)
            result += compound.ljust(20) + str(self.compound_masses[index])
            result += '\n'
        return result

    def __add__(self, other):
        """
        Add self and 'other' together, return the result as a new package, and
        leave self unchanged.

        :param other: Can can be one of the following:
          1. MaterialPackage: 'other' is added to self to create a new package.
          2. tuple: (compound, mass): The specified mass of the specified
          compound is added to self.

        :returns: A new Material package that is the sum of self and 'other'.
        """

        # Add another package.
        if type(other) is MaterialPackage:
            # Packages of the same material.
            if self.material == other.material:
                result = MaterialPackage(
                    self.material,
                    self.compound_masses + other.compound_masses)
                return result
            else:  # Packages of different materials.
                result = self.clone()
                for compound in other.material.compounds:
                    if compound not in self.material.compounds:
                        raise Exception(
                            "Packages of '" + other.material.name +
                            "' cannot be added to packages of '" +
                            self.material.name +
                            "'. The compound '" + compound +
                            "' was not found in '" + self.material.name + "'.")
                    result += (compound, other.get_compound_mass(compound))
                return result

        # Add the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            # Added material varialbes.
            compound = other[0]
            compound_index = self.material.get_compound_index(compound)
            mass = other[1]

            # Create the result package.
            result = self.clone()
            result.compound_masses[compound_index] += mass
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError('Invalid addition argument.')

    def __mul__(self, scalar):
        """
        The multiplication operator (*).

        Create a new package by multiplying self with scalar.

        :param scalar: The result is a new package with its content equal to
          self multiplied by a scalar, leaving self unchanged.

        :returns: A new MaterialPackage equal to self package multiplied by
          other.
        """

        # Multiply with a scalar floating point number.
        if type(scalar) is float:
            if scalar < 0.0:
                raise Exception(
                    'Invalid multiplication operation. '
                    'Cannot multiply package with negative number.')
            result = MaterialPackage(
                self.material, [c * scalar for c in self.compound_masses])
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError('Invalid multiplication argument.')

    def _validate_params_(self, material, compound_masses):
        if not type(material) is Material:
            raise TypeError('Invalid material type. Must be '
                            'chemistry.material.Material')
        if not type(compound_masses) is list:
            raise TypeError('Invalid compound_masses type. Must be '
                            'list.')

    def _is_compound_mass_tuple(self, value):
        """
        Determines whether value is a tuple of the format (compound(str),
        mass(float)).
        """

        if not type(value) is tuple:
            return False
        elif not len(value) == 2:
            return False
        elif not type(value[0]) is str:
            return False
        elif not type(value[1]) is float:
            return False
        else:
            return True

    def clone(self):
        """
        Create a complete copy of self.

        :returns: A MaterialPackage that is identical to self.
        """

        result = copy.copy(self)
        result.compound_masses = copy.deepcopy(self.compound_masses)

        return result

    # TODO: test
    def clear(self):
        """
        Set all the compound masses in the package to zero.
        """

        self.compound_masses *= 0.0

    def get_assay(self):
        """
        Determine the assay of self.

        :returns: [mass fractions] An array containing the assay of self.
        """

        masses_sum = sum(self.compound_masses)
        return [m / masses_sum for m in self.compound_masses]

    def get_mass(self):
        """
        Get the mass of the package.

        :returns: [kg]
        """

        return sum(self.compound_masses)

    def get_compound_mass(self, compound):
        """
        Get the mass of the specified compound in the package.

        :param compound: The formula of the compound, e.g. Fe2O3.

        :returns: [kg]
        """

        return self.compound_masses[self.material.get_compound_index(compound)]

    # TODO: Test
    def get_compound_mass_fraction(self, compound):
        """
        Get the mass fraction of the specified compound in self.

        :param compound: The formula and phase of the compound, e.g. Fe2O3.

        :returns: []
        """

        return self.get_compound_mass(compound) / self.get_mass()

    def get_element_masses(self):
        """
        Get the masses of elements in the package.

        :returns: [kg] An array of element masses. The sequence of the elements
          in the result corresponds with the sequence of elements in the
          element list of the material.
        """

        result = [0] * len(self.material.elements)
        for compound in self.material.compounds:
            c = self.get_compound_mass(compound)
            f = [c * x for x in emf(compound, self.material.elements)]
            result = [v+f[ix] for ix, v in enumerate(result)]

        return result

    def get_element_mass_dictionary(self):
        """
        Determine the masses of elements in the package and return as a
        dictionary.

        :returns: [kg] A dictionary of element symbols and masses.
        """

        element_symbols = self.material.elements
        element_masses = self.get_element_masses()

        result = {}
        for s, m in zip(element_symbols, element_masses):
            result[s] = m

        return result

    def get_element_mass(self, element):
        """
        Determine the masses of elements in the package.

        :returns: [kg] An array of element masses. The sequence of the elements
          in the result corresponds with the sequence of elements in the
          element list of the material.
        """

        result = [0]
        for compound in self.material.compounds:
            c = self.get_compound_mass(compound)
            f = [c * x for x in emf(compound, [element])]
            result = [v+f[ix] for ix, v in enumerate(result)]

        return result[0]

    def extract(self, other):
        """
        Extract 'other' from self, modifying self and returning the extracted
        material as a new package.

        :param other: Can be one of the following:

          * float: A mass equal to other is extracted from self. Self is
            reduced by other and the extracted package is returned as a
            new package.
          * tuple (compound, mass): The other tuple specifies the mass of
            a compound to be extracted. It is extracted from self and the
            extracted mass is returned as a new package.
          * string: The 'other' string specifies the compound to be extracted.
            All of the mass of that compound will be removed from self and a
            new package created with it.


        :returns: A new material package containing the material that was
          extracted from self.
        """

        # Extract the specified mass.
        if type(other) is float:

            if other > self.get_mass():
                raise Exception('Invalid extraction operation. Cannot extract'
                                'a mass larger than the package\'s mass.')

            fraction_to_subtract = other / self.get_mass()
            result = MaterialPackage(
                self.material,
                [m * fraction_to_subtract for m in self.compound_masses])
            self.compound_masses = [m * (1.0 - fraction_to_subtract)
                                    for m in self.compound_masses]

            return result

        # Extract the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            index = self.material.get_compound_index(other[0])

            if other[1] > self.compound_masses[index]:
                raise Exception('Invalid extraction operation. Cannot extract'
                                'a compound mass larger than what the package'
                                'contains.')

            self.compound_masses[index] -= other[1]
            resultarray = [0.0] * len(self.compound_masses)
            resultarray[index] = other[1]
            result = MaterialPackage(self.material, resultarray)

            return result

        # Extract all of the specified compound.
        elif type(other) is str:
            index = self.material.get_compound_index(other)
            result = self * 0.0
            result.compound_masses[index] = self.compound_masses[index]
            self.compound_masses[index] = 0.0

            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError('Invalid extraction argument.')

    # TODO: Test
    def add_to(self, other):
        """
        Add another chem material package to this material package.

        :param other: The other material package.
        """

        # Add another package.
        if type(other) is MaterialPackage:

            # Packages of the same material.
            if self.material == other.material:
                self.compound_masses += other.compound_masses

            # Packages of different materials.
            else:
                for compound in other.material.compounds:
                    if compound not in self.material.compounds:
                        raise Exception("Packages of '" + other.material.name +
                                        "' cannot be added to packages of '" +
                                        self.material.name +
                                        "'. The compound '" + compound +
                                        "' was not found in '" +
                                        self.material.name + "'.")
                    self.add_to((compound, other.get_compound_mass(compound)))

        # Add the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            # Added material varialbes.
            compound = other[0]
            compound_index = self.material.get_compound_index(compound)
            mass = other[1]

            # Create the result package.
            self.compound_masses[compound_index] += mass

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError('Invalid addition argument.')

if __name__ == "__main__":
    import unittest
    from auxi.modelling.process.materials.chem_test \
        import ChemMaterialUnitTester, ChemMaterialPackageUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides a set of physical constants that are used frequently.
"""

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


# TODO: Add literature references to these constants.

# Universal Constants

c = 299792458
"""[m·s-1] speed of light in vacuum"""

G = 6.67408E-11
"""[m3.kg-1.s-2] Newtonian constant of gravitation"""

h = 6.626070040E-34
"""[J.s] Planck constant"""

h_bar = 1.054571800E-34
"""[J.s] reduced Planck constant (h/(2pi))"""


# Physico-chemical Constants

m_u = 1.660538921E-27
"""[kg] atomic mass constant"""

N_A = 6.02214129E23
"""[mol-1] Avogadro's number"""

k_B = 1.3806488E-23
"""[J.K-1] Boltzmann constant"""

F = 96485.3365
"""[C·mol-1] Faraday constant"""

R = 8.3144621
"""[J.K-1.mol-1] gas constant"""


# Other Constants

g = 9.80665
"""[m.s-2] standard acceleration of gravity"""



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest

from auxi.tools.chemistry import stoichiometry as testee


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class StoichFunctionTester(unittest.TestCase):
    """
    The function tester for the stoichiometry module.
    """

    def assertAlmostEqual(self, first, second, places=14, msg=None,
                          delta=None):
        if type(first) is list and type(second) is list:
            self.assertEqual(len(first), len(second))
            for f, s in zip(first, second):
                self.assertAlmostEqual(f, s)
        else:
            super(StoichFunctionTester, self).assertAlmostEqual(
                first, second, places, msg, delta)

    def test_invalid_characters(self):
        """
        Test whether an exception is raised when a compound formula contains an
        invalid character.
        """

        self.assertRaises(ValueError, testee.molar_mass, '(FeO)*(Fe2O3)')
        self.assertRaises(ValueError, testee.molar_mass, 'CaO1,5')
        self.assertRaises(ValueError, testee.stoichiometry_coefficients,
                          'CaO1,5', ['Ca', 'O'])

    def test_molar_mass(self):
        """
        Test whether the molar mass of a compound is calculated correctly.
        """

        func = testee.molar_mass

        self.assertAlmostEqual(func('FeO'), 71.8444)
        self.assertAlmostEqual(func('Fe2O3'), 159.6882)
        self.assertAlmostEqual(func('SiO2'), 60.0843)
        self.assertAlmostEqual(func('Ca(OH)2'), 74.09268)

    def test_amount(self):
        """
        Test whether the amount of a compound is calculated correctly.
        """

        func = testee.amount

        self.assertEqual(func('FeO', 1.0), 0.01391896932815919)

    def test_mass(self):
        """
        Test whether the mass of a compound is calculated correctly.
        """

        func = testee.mass

        self.assertEqual(func('FeO', 1.0), 71.8444)

    def test_convert_compound(self):
        """
        Test whether compound conversions are calculated correctly.
        """

        func = testee.convert_compound

        m_Fe = 1000.0
        m_Fe2O3 = func(m_Fe, 'Fe', 'Fe2O3', 'Fe')
        self.assertAlmostEqual(m_Fe2O3, 1429.7448294386247)
        m_Fe2O3 = func(m_Fe, 'Fe', 'Fe2O3', 'O')
        self.assertEqual(m_Fe2O3, 0.0)

        m_FeTiO3 = 1000.0
        m_FeO = func(m_FeTiO3, 'FeTiO3', 'FeO', 'Fe')
        self.assertAlmostEqual(m_FeO, 473.56341234801613)
        m_TiO2 = func(m_FeTiO3, 'FeTiO3', 'TiO2', 'Ti')
        self.assertAlmostEqual(m_TiO2, 526.4365876519838)
        self.assertAlmostEqual(m_FeO + m_TiO2, m_FeTiO3)

    def test_element_mass_fraction(self):
        """
        Test whether an element mass fraction is calculated correctly.
        """

        func = testee.element_mass_fraction

        self.assertAlmostEqual(func('FeO', 'Fe'), 0.7773048421310499)
        self.assertAlmostEqual(func('FeO', 'O'), 0.22269515786895014)
        self.assertAlmostEqual(func('FeO', 'Si'), 0.0)
        self.assertAlmostEqual(func('Fe2O3', 'Fe'), 0.699425505453753)
        self.assertAlmostEqual(func('Fe2O3', 'O'), 0.300574494546247)
        self.assertAlmostEqual(func('Fe2O3', 'Mn'), 0.0)
        self.assertAlmostEqual(func('SiO2', 'Si'), 0.4674349206032192)
        self.assertAlmostEqual(func('SiO2', 'O'), 0.5325650793967809)
        self.assertAlmostEqual(func('SiO2', 'Fe'), 0.0)
        self.assertAlmostEqual(func('Ca(OH)2', 'Ca'), 0.5409171324346751)
        self.assertAlmostEqual(func('Ca(OH)2', 'O'), 0.43187532155673136)
        self.assertAlmostEqual(func('Ca(OH)2', 'H'), 0.027207546008593562)
        self.assertAlmostEqual(func('Ca(OH)2', 'Fe'), 0.0)

    def test_element_mass_fractions(self):
        """
        Test whether a list of element mass fraction is calculated correctly.
        """

        func = testee.element_mass_fractions

        x_FeO = [0.7773048421310499, 0.22269515786895014, 0.0]
        self.assertAlmostEqual(list(func('FeO', ['Fe', 'O', 'Si'])), x_FeO)

        x_Fe2O3 = [0.699425505453753, 0.300574494546247, 0.0]
        self.assertAlmostEqual(list(func('Fe2O3', ['Fe', 'O', 'Mn'])), x_Fe2O3)

        x_SiO2 = [0.4674349206032192, 0.5325650793967809, 0.0]
        self.assertAlmostEqual(list(func('SiO2', ['Si', 'O', 'Fe'])), x_SiO2)

        x_CaO2H2 = [0.5409171324346751, 0.43187532155673136,
                    0.027207546008593562, 0.0]
        self.assertAlmostEqual(list(func('Ca(OH)2', ['Ca', 'O', 'H', 'Fe'])),
                               x_CaO2H2)

    def test_elements(self):
        """
        Test whether the set of elements in a list of compounds is calculated
        correctly.
        """

        func = testee.elements

        compounds = ['Ar', 'Fe2O3', 'SiO2', 'Al2O3', 'SO3', 'CaO', 'Fe',
                     'Mn2O3']
        elements = func(compounds)
        self.assertEqual(elements, {'Al', 'Ar', 'Ca', 'Fe', 'Mn', 'O', 'S',
                                    'Si'})

        compounds = ['CO2']
        elements = func(compounds)
        self.assertEqual(elements, {'O', 'C'})

    def test_stoichiometry_coefficient(self):
        """
        Test whether the stoichiometry coefficient of a specified element in a
        compound is calculated correctly.
        """

        func = testee.stoichiometry_coefficient

        self.assertAlmostEqual(func('FeO', 'Fe'), 1.0)
        self.assertAlmostEqual(func('FeO', 'O'), 1.0)
        self.assertAlmostEqual(func('FeO', 'Si'), 0.0)
        self.assertAlmostEqual(func('Fe2O3', 'Fe'), 2.0)
        self.assertAlmostEqual(func('Fe2O3', 'O'), 3.0)
        self.assertAlmostEqual(func('Fe2O3', 'Mn'), 0.0)
        self.assertAlmostEqual(func('SiO2', 'Si'), 1.0)
        self.assertAlmostEqual(func('SiO2', 'O'), 2.0)
        self.assertAlmostEqual(func('SiO2', 'Fe'), 0.0)
        self.assertAlmostEqual(func('Ca(OH)2', 'Ca'), 1.0)
        self.assertAlmostEqual(func('Ca(OH)2', 'O'), 2.0)
        self.assertAlmostEqual(func('Ca(OH)2', 'H'), 2.0)
        self.assertAlmostEqual(func('Ca(OH)2', 'Fe'), 0.0)

    def test_stoichiometry_coefficients(self):
        """
        Test whether the stoichiometry coefficients of a specified list of
        elements in a compound is calculated correctly.
        """

        func = testee.stoichiometry_coefficients

        self.assertEqual(func('FeO', ['Fe', 'O', 'Si']),
                         [1.0, 1.0, 0.0])
        self.assertEqual(func('Fe2O3', ['Fe', 'O', 'Mn']),
                         [2.0, 3.0, 0.0])
        self.assertEqual(func('FeOFe2O3', ['Fe', 'O', 'Mn']),
                         [3.0, 4.0, 0.0])
        self.assertEqual(func('(FeO)(Fe2O3)', ['Fe', 'O', 'Mn']),
                         [3.0, 4.0, 0.0])
        self.assertEqual(func('SiO2', ['Si', 'O', 'Fe']),
                         [1.0, 2.0, 0.0])
        self.assertEqual(func('Ca(OH)2', ['Ca', 'O', 'H', 'Fe']),
                         [1.0, 2.0, 2.0, 0.0])
        self.assertEqual(func('CaAl2(Si2O7)(OH)2.H2O',
                              ['Al', 'Ca', 'Si', 'O', 'H']),
                         [2.0, 1.0, 2.0, 10.0, 4.0])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for the
auxi.core.time module.
"""

import unittest
from datetime import datetime

from auxi.core.time import Clock, TimePeriod

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ClockUnitTester(unittest.TestCase):
    """
    Tester for the auxi.core.time.Clock class.
    """

    def setUp(self):
        self.object = Clock("NameA",
                            description="DescriptionA",
                            start_datetime=datetime(2016, 2, 1),
                            timestep_period_duration=TimePeriod.year,
                            timestep_period_count=3)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.timestep_period_duration, TimePeriod.year)
        self.assertEqual(self.object.timestep_period_count, 3)

    def test_tick(self):
        self.assertEqual(self.object.timestep_ix, 0)
        self.object.tick()
        self.assertEqual(self.object.timestep_ix, 1)

    def test_reset(self):
        self.object.tick()
        self.object.reset()
        self.assertEqual(self.object.timestep_ix, 0)

    def test_get_datetime_at_period_ix(self):
        self.assertEqual(self.object.get_datetime_at_period_ix(3),
                         datetime(2019, 2, 1))

    def test_get_datetime(self):
        self.object.tick()
        self.assertEqual(self.object.get_datetime(), datetime(2017, 2, 1))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
from os.path import realpath, dirname

from auxi.tools.transportphenomena.heattransfer import naturalconvection as testee
from auxi.tools.materialphysicalproperties.gases import air


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


MODULE_PATH = dirname(realpath(__file__))


class IsothermalFlatSurface_RegionTester(unittest.TestCase):

    def setUp(self):
        self.region = testee.IsothermalFlatSurface.Region(
            "r1", -90, True, -60, True, None, False, 4, False)

    def test_constructor(self):
        self.assertEqual(self.region.theta_min, -90)
        self.assertTrue(self.region.theta_min_incl, True)
        self.assertEqual(self.region.theta_max, -60)
        self.assertTrue(self.region.theta_max_incl)
        self.assertIsNone(self.region.Ra_min)
        self.assertFalse(self.region.Ra_min_incl)
        self.assertEqual(self.region.Ra_max, 4)
        self.assertFalse(self.region.Ra_max_incl)

    def test_contains_point(self):
        # testing theta < min boundary
        self.assertFalse(self.region.contains_point(-91, 3))
        # testing theta > max boundary
        self.assertFalse(self.region.contains_point(-59, 3))
        # testing Ra > max boundary
        self.assertFalse(self.region.contains_point(-75, 5))
        # testing all within boundary
        self.assertTrue(self.region.contains_point(-75, 3))

    @unittest.skip("Todo: Implement Test. I do not know how to.")
    def test_plot_region(self):
        self.assertFalse(True)


class IsothermalFlatSurfaceTester(unittest.TestCase):

    def setUp(self):
        self.model = testee.IsothermalFlatSurface(air)

        self.L = 0.4
        self.theta = 0.0
        self.Ts = 313.0
        self.Tf = 283.0

    def test_constructor(self):
        self.assertEqual(self.model.references, None)
        self.assertEqual(len(self.model.equation_dict), 22)
        self.assertEqual(len(self.model.regions), 22)

    def test_Nu_x(self):
        Nu_x = self.model.Nu_x(self.L, self.theta, self.Ts, T=self.Tf)
        self.assertAlmostEqual(Nu_x, 64.1972833727)

    def test_Nu_L(self):
        Nu_L = self.model.Nu_L(self.L, self.theta, self.Ts, T=self.Tf)
        self.assertAlmostEqual(Nu_L, 85.5963778303)

    def test_h_x(self):
        h_x = self.model.h_x(self.L, self.theta, self.Ts, T=self.Tf)
        self.assertAlmostEqual(h_x, 4.21863186596)

    def test_h_L(self):
        h_L = self.model.h_L(self.L, self.theta, self.Ts, T=self.Tf)
        self.assertAlmostEqual(h_L, 5.62484248795)

    def test_plot_regions(self):
        pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for classes in the chem module.
"""

import unittest

from auxi.modelling.process.materials.core import Material
from auxi.tools.materialphysicalproperties.gases import air_dict


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class MaterialTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.process.materials.core.Material class.
    """

    def test_constructor(self):
        Material('Test Material', air_dict, 'Test material.')

    def test_beta(self):
        m = Material('Test Material', air_dict, 'Test material.')

        T = 300.0
        self.assertEqual(m.beta(T=T), 1.0/T)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides class and functions for basic business activities.
"""

from datetime import datetime
from dateutil.relativedelta import relativedelta

from auxi.modelling.business.structure import Activity

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class BasicActivity(Activity):
    """
    An activity class that provides the most basic activity functionality:
    periodically create a transaction between two specified accounts.

    :param name: The name.
    :param dt_account: The debit account.
    :param cr_account: The credit account.
    :param amount: The amount of an activity.
    :param start: The datetime the activity should be started.
    :param end: The datetime the activity should be run until.
    :param interval: The interval of the activity.
    :param description: The description.
    """

    def __init__(self, name,
                 dt_account,
                 cr_account,
                 amount=0,
                 start=datetime.min,
                 end=datetime.max,
                 interval=1,
                 description=None):
        """
        """
        super(BasicActivity, self).__init__(
            name,
            description=description,
            start=start,
            end=end,
            interval=interval)
        self.interval = interval
        self.amount = amount
        self.dt_account = dt_account
        self.cr_account = cr_account

    def _meet_execution_criteria(self, ix_period):
        return super(BasicActivity, self)._meet_execution_criteria(
            ix_period) and self.amount > 0

    def run(self, clock, generalLedger):
        """
        Execute the activity at the current clock cycle.

        :param clock: The clock containing the current execution time and
          period information.
        :param generalLedger: The general ledger into which to create the
          transactions.
        """

        if not self._meet_execution_criteria(clock.timestep_ix):
            return

        generalLedger.create_transaction(
            self.description if self.description is not None else self.name,
            description='',
            tx_date=clock.get_datetime(),
            dt_account=self.dt_account,
            cr_account=self.cr_account,
            source=self.path,
            amount=self.amount)

    def get_referenced_accounts(self):
        """
        Retrieve the general ledger accounts referenced in this instance.

        :returns: The referenced accounts.
        """

        return [self.dt_account, self.cr_account]

'''
class BasicEquipment(Activity):
    """
    An activity class that provides a basic activity functionality for
    a purchase of equipment. This includes the purchase and depreciation
    thereafter.

    :param name: The name.
    :param description: The description.
    :param bank_account: The asset account that is decreased.
    :param equipment_account: The fixed asset account that is increased.
    :param depreciation_account: The expense account that is increased.
    :param purchase_amount: The purchase amount. The default amount is 0.
    :param writeoff_amount: The writeoff amount. The default amount is 0.
    :param depreciation_rate: The rate of depreciation as fraction of the whole
      (e.g. 0.15 = 15%). The default value is 0.0.
    :param start: The datetime the activity should be started.
    :param lifetime: The amount of months the equipment can exists before
      writing it off.
    :param interval: The interval of the activity.
    """

    def __init__(self, name,
                 bank_account,
                 loan_account,
                 interest_account,
                 purchase_amount=0,
                 writeoff_amount=0,
                 depreciation_rate=0.0,
                 start=datetime.min,
                 lifetime=60,
                 interval=1,
                 description=None):
        """
        """
        super(BasicEquipment, self).__init__(
            name,
            description=description,
            start=start,
            end=start + relativedelta(months=duration),
            interval=interval)
        self.bank_account = bank_account
        self.loan_account = loan_account
        self.interest_account = interest_account
        self._amount = amount
        self.duration = duration
        self.interest_rate = interest_rate

        self._months_executed = 0
        self._amount_left = amount
'''


class BasicLoanActivity(Activity):
    """
    An activity class that provides the most basic activity functionality for
    a loan:
    Creates a loan transaction and periodically create transactions to consider
    the interest and to pay the interest.

    :param name: The name.
    :param description: The description.
    :param bank_account: The asset account that is increased.
    :param loan_account: The liability account that is decreased.
    :param interest_account: The expense account the interest is
      added to.
    :param amount: The loan amount. The default amount is 0
    :param interest_rate: The interest rate as a fraction of the whole
      (e.g. 0.15 = 15%). The default value is 0.0
    :param start: The datetime the activity should be started.
    :param duration: The duration of the loan in months.
    :param interval: The interval of the activity.
    """

    def __init__(self, name,
                 bank_account,
                 loan_account,
                 interest_account,
                 amount=0,
                 interest_rate=0.0,
                 start=datetime.min,
                 duration=60,
                 interval=1,
                 description=None):
        """
        """
        super(BasicLoanActivity, self).__init__(
            name,
            description=description,
            start=start,
            end=start + relativedelta(months=duration+1),
            interval=interval)
        self.bank_account = bank_account
        self.loan_account = loan_account
        self.interest_account = interest_account
        self._amount = amount
        self.duration = duration
        self.interest_rate = interest_rate

        self._months_executed = 0
        self._amount_left = amount

    def _update_montly_payment_(self):
        int_rate = self._montly_interest_rate
        val = (self.amount * int_rate)
        val = val / (1 - (1 / pow((1 + int_rate), self.duration)))
        self._monthly_payment = val

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, value):
        self._amount = value
        self._update_montly_payment_()

    @property
    def interest_rate(self):
        return self._interest_rate

    @interest_rate.setter
    def interest_rate(self, value):
        self._interest_rate = value
        self._montly_interest_rate = value / 12.0
        self._update_montly_payment_()

    def _meet_execution_criteria(self, ix_period):
        return super(BasicLoanActivity, self)._meet_execution_criteria(
            ix_period) and self.amount > 0 and self.duration > 0

    def prepare_to_run(self, clock, period_count):
        """
        Prepare the activity for execution.

        :param clock: The clock containing the execution start time and
          execution period information.
        :param period_count: The total amount of periods this activity will be
          requested to be run for.
        """

        super(BasicLoanActivity, self).prepare_to_run(clock, period_count)

        self._months_executed = 0
        self._amount_left = self.amount

    def run(self, clock, generalLedger):
        """
        Execute the activity at the current clock cycle.

        :param clock: The clock containing the current execution time and
          period information.
        :param generalLedger: The general ledger into which to create the
          transactions.
        """

        if not self._meet_execution_criteria(clock.timestep_ix):
            return

        if self.description is None:
            tx_name = self.name
        else:
            tx_name = self.description

        if self._months_executed == 0:
            generalLedger.create_transaction(
                tx_name,
                description='Make a loan',
                tx_date=clock.get_datetime(),
                dt_account=self.bank_account,
                cr_account=self.loan_account,
                source=self.path,
                amount=self.amount)
        else:
            curr_interest_amount = (self._amount_left *
                                    self.interest_rate) / 12.0

            generalLedger.create_transaction(
                tx_name,
                description='Consider interest',
                tx_date=clock.get_datetime(),
                dt_account=self.interest_account,
                cr_account=self.loan_account,
                source=self.path,
                amount=curr_interest_amount)

            generalLedger.create_transaction(
                tx_name,
                description='Pay principle',
                tx_date=clock.get_datetime(),
                dt_account=self.loan_account,
                cr_account=self.bank_account,
                source=self.path,
                amount=self._monthly_payment)

            self._amount_left += curr_interest_amount - self._monthly_payment

        self._months_executed += self.interval

    def get_referenced_accounts(self):
        """
        Retrieve the general ledger accounts referenced in this instance.

        :returns: The referenced accounts.
        """

        return [
            self.bank_account,
            self.loan_account,
            self.interest_account]


if __name__ == "__main__":
    import unittest
    from auxi.modelling.business.basic_test import BasicActivityUnitTester
    from auxi.modelling.business.basic_test import BasicLoanActivityUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides tools for calculating material physical properties.
"""


import csv
import os
import pandas as pd
import webbrowser
from cerberus import Validator, ValidationError
from enum import Enum

from auxi.core.objects import Object


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Marno Grewar, Christoff Kok, Johan Zietsman'
__credits__ = ['Marno Grewar', 'Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Johan Zietsman'
__email__ = 'johan.zietsman@ex-mente.co.za'
__status__ = 'Planning'


class StateOfMatter(Enum):
    unknown = 0
    solid = 1
    liquid = 2
    solid_liquid = solid | liquid
    gas = 4
    solid_gas = solid | gas
    solid_liquid_gas = solid | liquid | gas
    plasma = 8
    solid_plasma = solid | plasma
    solid_liquid_plasma = solid | liquid | plasma
    solid_liquid_gas_plasma = solid | liquid | gas | plasma
    liquid_plasma = liquid | plasma
    liquid_gas_plasma = liquid | gas | plasma
    gas_plasma = gas | plasma


class DataSet(object):
    """
    Contains a data set used to create and/or test a material physical property
    model.

    :param csvfilepath: path to the csv file that contains the data
    """

    def create_template(material, path, show=False):
        """
        Create a template csv file for a data set.

        :param material: the name of the material
        :param path: the path of the directory where the file must be written
        :param show: a boolean indicating whether the created file should be \
        displayed after creation
        """
        file_name = 'dataset-%s.csv' % material.lower()
        file_path = os.path.join(path, file_name)

        with open(file_path, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',',
                                quotechar='"', quoting=csv.QUOTE_MINIMAL)
            writer.writerow(['Name', material])
            writer.writerow(['Description', '<Add a data set description '
                                            'here.>'])
            writer.writerow(['Reference', '<Add a reference to the source of '
                                          'the data set here.>'])
            writer.writerow(['Temperature', '<parameter 1 name>',
                            '<parameter 2 name>', '<parameter 3 name>'])
            writer.writerow(['T', '<parameter 1 display symbol>',
                             '<parameter 2 display symbol>',
                             '<parameter 3 display symbol>'])
            writer.writerow(['K', '<parameter 1 units>',
                             '<parameter 2 units>', '<parameter 3 units>'])
            writer.writerow(['T', '<parameter 1 symbol>',
                             '<parameter 2 symbol>', '<parameter 3 symbol>'])
            for i in range(10):
                writer.writerow([100.0 + i*50, float(i), 10.0 + i, 100.0 + i])

        if show is True:
            webbrowser.open_new(file_path)

    def __init__(self, csvfilepath):
        self._file_path = csvfilepath
        self._read_header_information()
        self._read_data()

    def __str__(self):
        result = 'Material:         %s\n' % self.material
        result += 'Description:      %s\n' % self.description
        result += 'References:       %s\n' % self.reference
        result += 'Properties:       %s\n' % self.col_names
        result += 'Symbols:          %s\n' % self.col_symbols
        result += 'Display symbols:  %s\n' % self.col_display_symbols
        result += 'Units:            %s\n' % self.col_units
        result += 'Data:\n'
        result += str(self.data)

        return result

    def _read_header_information(self):
        with open(self._file_path, newline='') as file:
            lines = csv.reader(file, delimiter=',', quotechar='"')

            self.name = os.path.basename(self._file_path)
            """The name of the data set."""

            self.material = next(lines)[1]
            """The name of the material represented by the data set."""

            self.description = next(lines)[1]
            """A description of the data set."""

            self.reference = next(lines)[1]
            """A reference to the source of the data set."""

            self.col_names = next(lines)
            """A list of data set column names."""

            self.col_display_symbols = next(lines)
            """A list of data set column display symbols."""

            self.col_units = next(lines)
            """A list of data set column units."""

            self.col_symbols = next(lines)
            """A list of data set column symbols."""

            self.names_dict = dict(zip(self.col_symbols, self.col_names))
            """A dictionary to translate a parameter's symbol to its name."""

            self.display_symbols_dict = dict(zip(self.col_symbols,
                                                 self.col_display_symbols))
            """
            A dictionary to translate a parameter's symbol to its display
            symbol.
            """

            self.units_dict = dict(zip(self.col_symbols, self.col_units))
            """A dictionary to translate a parameter's symbol to its units."""

    def _read_data(self):
        self.data = pd.read_csv(self._file_path, header=6)


class Model(Object):
    """
    Base class of models that describe the variation of a specific material
    physical property.

    :param material: the name of the material being described, e.g. "Air"
    :param proprty: the name of the property being described, e.g. "density"
    :param symbol: the symbol of the property being described, e.g. "rho"
    :param display symbol: the display symbol of the property being described,\
    e.g. "\rho"
    :param units: the units used to express the property, e.g. "kg/m3"
    :param references: a list of literature references on which this model\
    implementation is based, e.g. ['lienhard2015', 'smith2006']
    :param datasets: a list of data sets on which this model implementation is\
    based, e.g. ['dataset-air-lienhard2015.csv']
    """

    def __init__(self, material, proprty, symbol, display_symbol, units,
                 state_schema, references, datasets):
        self.material = material
        self.property = proprty
        self.symbol = symbol
        self.display_symbol = display_symbol
        self.units = units
        self.references = references
        self.datasets = datasets
        self.state_schema = state_schema
        self._create_validator(self.state_schema)

    def __call__(self, **state):
        """
        For details about this method, please refer to the calculate method.
        """
        return self.calculate(**state)

    def _create_validator(self, state_schema):
        self.state_schema = state_schema
        self.state_validator = Validator(self.state_schema)

    def calculate(self, **state):
        """
        Base calculate method for models.
        Validates the material state parameter(s).

        :param **state: The material state
        """
        if not self.state_validator.validate(state):
            msg = 'The state description contains errors:'
            for key, value in self.state_validator.errors.items():
                msg += ' %s: %s;' % (key, value)
            msg = msg[0:-1]+'.'
            raise ValidationError(msg)


if __name__ == "__main__":
    import unittest
    from core_test import *
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for classes in the chem module.
"""

import unittest

from auxi.core.helpers import get_path_relative_to_module as get_path
from auxi.modelling.process.materials.chem import Material, MaterialPackage


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ChemMaterialUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.process.materials.chem.Material class.
    """

    def assertAlmostEqual(self, first, second, places=14, msg=None,
                          delta=None):
        if type(first) is list and type(second) is list:
            self.assertEqual(len(first), len(second))
            for f, s in zip(first, second):
                self.assertAlmostEqual(f, s)
        else:
            super(ChemMaterialUnitTester, self).assertAlmostEqual(
                first, second, places, msg, delta)

    def setUp(self):
        test_data_file_path = get_path(
            __file__, 'data/chemmaterial.test.ilmenite.txt')
        self.material = Material("material", test_data_file_path)

    def test_constructor(self):
        self.assertEqual(self.material.name, "material")
        self.assertEqual(len(self.material.compounds), 14)
        self.assertEqual(self.material.compound_count, 14)
        self.assertEqual(len(self.material.assays), 3)

    def test_get_compound_index(self):
        self.assertEqual(self.material.get_compound_index("Al2O3"), 0)
        self.assertEqual(self.material.get_compound_index("K2O"), 6)
        self.assertEqual(self.material.get_compound_index("P4O10"), 10)
        self.assertEqual(self.material.get_compound_index("V2O5"), 13)

    def test_create_empty_assay(self):
        empty_assay = self.material.create_empty_assay()
        self.assertEqual(len(empty_assay), 14)
        self.assertEqual(sum(empty_assay), 0.0)
        self.assertEqual(sum(empty_assay), 0.0)
        self.assertEqual(sum(empty_assay), 0.0)

    def test_add_assay(self):
        new_assay = self.material.create_empty_assay()
        new_assay[0] = 0.5
        new_assay[2] = 0.5
        self.material.add_assay("new_assay", new_assay)
        self.assertTrue(self.material.assays["new_assay"] == new_assay)

    def test_get_assay_total(self):
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteA"),
                               0.99791999999999992)
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteB"),
                               0.99760999999999989)
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteC"),
                               1.0000200000000001)

    def test_create_package(self):
        package = self.material.create_package("IlmeniteA", 123.456, True)
        self.assertEqual(package.get_mass(), 123.456)


class ChemMaterialPackageUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.process.materials.chemistry.material.Material
    class.
    """

    def setUp(self):
        path = get_path(__file__, 'data/chemmaterial.test.ilmenite.txt')
        self.ilm = Material("ilmenite", path)

        self.ilm_pkg_a = self.ilm.create_package("IlmeniteA", 1234.5, True)
        self.ilm_pkg_b = self.ilm.create_package("IlmeniteB", 2345.6, True)
        self.ilm_pkg_c = self.ilm.create_package("IlmeniteC", 3456.7, True)

        path = get_path(__file__, 'data/chemmaterial.test.reductant.txt')
        self.red = Material("reductant", path)

        self.red_pkg_a = self.red.create_package("ReductantA", 123.45, True)

        path = get_path(__file__, 'data/chemmaterial.test.mix.txt')
        self.mix = Material("mix", path)

    def test_constructor(self):
        m_sum = sum(self.ilm.assays["IlmeniteB"])
        compound_masses = [(m * 123.4) / m_sum
                           for m in self.ilm.assays["IlmeniteB"]]
        package = MaterialPackage(self.ilm, compound_masses)
        self.assertAlmostEqual(package.get_mass(), 123.4)

    def test_add_operator_1(self):
        """
        other = MaterialPackage

        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        pkg_a_plus_b = self.ilm_pkg_a + self.ilm_pkg_b
        pkg_a_plus_c = self.ilm_pkg_a + self.ilm_pkg_c
        pkg_b_plus_c = self.ilm_pkg_b + self.ilm_pkg_c
        pkg_a_plus_b_plus_c = self.ilm_pkg_a + self.ilm_pkg_b + \
            self.ilm_pkg_c

        self.assertAlmostEqual(self.ilm_pkg_a.get_mass(), 1234.5)

        self.assertAlmostEqual(self.ilm_pkg_b.get_mass(), 2345.6)

        self.assertAlmostEqual(self.ilm_pkg_c.get_mass(), 3456.7)

        self.assertAlmostEqual(pkg_a_plus_b.get_mass(), 3580.1)

        self.assertAlmostEqual(pkg_a_plus_c.get_mass(), 4691.2)

        self.assertAlmostEqual(pkg_b_plus_c.get_mass(), 5802.3)

        self.assertAlmostEqual(pkg_a_plus_b_plus_c.get_mass(), 7036.8)

    def test_add_operator_2(self):
        """
        Tests the scenario when a 'mix' is created and two packages
        'mixed' into the mix package.
        """
        mix_package = self.mix.create_package(None, 0.0)
        mix_package = mix_package + self.ilm_pkg_a
        mix_package = mix_package + self.red_pkg_a

        self.assertAlmostEqual(mix_package.get_mass(),
                               self.ilm_pkg_a.get_mass() +
                               self.red_pkg_a.get_mass(),
                               places=10)

        self.assertRaises(Exception, self.add_incompatible_packages)

    def add_incompatible_packages(self):
        result = self.ilm_pkg_a + self.red_pkg_a
        result = result * 1.0

    def test_add_operator_3(self):
        """other = tuple (compound, mass)
        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        packageAplusAl2O3 = self.ilm_pkg_a + ("Al2O3", 123.4)

        self.assertEqual(packageAplusAl2O3.get_mass(), 1357.9)

    def test_extract_1(self):
        temp_package_a = self.ilm_pkg_a.clone()
        mass = 432.1
        diff_package = temp_package_a.extract(mass)

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilm_pkg_a.get_mass() - mass)

        self.assertAlmostEqual(diff_package.get_mass(), mass)

    def test_subtract_operator_2(self):
        temp_package_a = self.ilm_pkg_a.clone()
        compound = "TiO2"
        mass = 123.4
        diff_package = temp_package_a.extract((compound, mass))

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilm_pkg_a.get_mass() - mass)
        self.assertAlmostEqual(temp_package_a.get_compound_mass(compound),
                               self.ilm_pkg_a.get_compound_mass(compound) -
                               mass)

        self.assertEqual(diff_package.get_mass(), mass)
        self.assertEqual(diff_package.get_compound_mass(compound), mass)

    def test_subtract_operator_3(self):
        temp_package_a = self.ilm_pkg_a.clone()
        compound = "TiO2"
        mass = temp_package_a.get_compound_mass(compound)
        diff_package = temp_package_a.extract(compound)

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilm_pkg_a.get_mass() - mass)
        self.assertAlmostEqual(temp_package_a.get_compound_mass(compound),
                               self.ilm_pkg_a.get_compound_mass(compound) -
                               mass)

        self.assertEqual(diff_package.get_mass(), mass)
        self.assertEqual(diff_package.get_compound_mass(compound), mass)

    def test_multiply_operator(self):
        temp_package_a = self.ilm_pkg_a.clone()

        mul_package_1 = temp_package_a * 0.0
        self.assertEqual(mul_package_1.get_mass(), 0.0)
        self.assertTrue(mul_package_1.compound_masses == [0.0] *
                        len(temp_package_a.compound_masses))

        mul_package_2 = temp_package_a * 1.0
        self.assertAlmostEqual(mul_package_2.get_mass(),
                               temp_package_a.get_mass(),
                               places=10)
        self.assertTrue(mul_package_2.compound_masses ==
                        temp_package_a.compound_masses)

        mul_package_2 = temp_package_a * 123.4
        self.assertAlmostEqual(mul_package_2.get_mass(),
                               temp_package_a.get_mass() * 123.4)
        self.assertTrue(mul_package_2.compound_masses ==
                        [m * 123.4 for m in temp_package_a.compound_masses])

    def test_clone(self):
        clone = self.ilm_pkg_a.clone()

        self.assertEqual(clone.get_mass(), self.ilm_pkg_a.get_mass())
        self.assertTrue(clone.compound_masses ==
                        self.ilm_pkg_a.compound_masses)

    def test_get_mass(self):
        self.assertAlmostEqual(self.ilm_pkg_a.get_mass(), 1234.5)
        self.assertAlmostEqual(self.ilm_pkg_b.get_mass(), 2345.6)
        self.assertAlmostEqual(self.ilm_pkg_c.get_mass(), 3456.7)

    def test_get_assay(self):
        ass = self.ilm.assays["IlmeniteA"]
        ass_sum = sum(ass)
        for ix, val in enumerate(self.ilm_pkg_a.get_assay()):
            self.assertTrue(val-(ass[ix] / ass_sum) < 1.0E-16)
            self.assertTrue(val-(ass[ix] / ass_sum) > -1.0E-16)

    def test_get_compound_mass(self):
        assay = "IlmeniteA"
        compound = "TiO2"
        for compound in self.ilm.compounds:
            index = self.ilm.get_compound_index(compound)
            mass = 1234.5 * self.ilm.assays[assay][index] / \
                self.ilm.get_assay_total(assay)
            self.assertEqual(
                self.ilm_pkg_a.get_compound_mass(compound),
                mass)

    def test_get_element_masses(self):
        x = self.ilm_pkg_a.get_element_masses()
        y = self.ilm_pkg_a.get_element_mass("Ti")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides a material class that can do thermochemical calculations.
"""

from auxi.core.objects import NamedObject
from auxi.tools.materialphysicalproperties.core import StateOfMatter


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Material(NamedObject):
    """
    Represents a material consisting of multiple chemical compounds, having
    the ability to do thermochemical calculations.

    :param name: A name for the material.
    :param state_of_matter: The material's state of matter, e.g. liquid.
    :param file_path: The location of the file containing the material's data.
    :param description: the material's description
    """

    def __init__(self, name, state_of_matter=StateOfMatter.unknown,
                 property_models=None, description=None):
        super().__init__(name, description)

        self.mstate = state_of_matter
        """State of matter."""

        if 'alpha' in property_models:
            self.alpha = property_models['alpha']

        if 'beta' in property_models:
            self.beta = property_models['beta']

        if 'Cp' in property_models:
            self.Cp = property_models['Cp']

        if 'k' in property_models:
            self.k = property_models['k']

        if 'mu' in property_models:
            self.mu = property_models['mu']

        if 'nu' in property_models:
            self.nu = property_models['nu']

        if 'rho' in property_models:
            self.rho = property_models['rho']

    def alpha(self, **state):
        """
        Calculate the alpha value given the material state.

        :param **state: material state

        :returns: float
        """

        return self.k(**state) / self.rho(**state) / self.Cp(**state)

    def beta(self, **state):
        """
        Calculate the alpha value given the material state.

        :param **state: material state

        :returns: float
        """
        raise NotImplementedError()

    def Cp(self, **state):
        """
        Calculate the Cp value given the material state.

        :param **state: material state

        :returns: float
        """
        raise NotImplementedError()

    def k(self, **state):
        """
        Calculate the k value given the material state.

        :param **state: material state

        :returns: float
        """
        raise NotImplementedError()

    def mu(self, **state):
        """
        Calculate the mu value given the material state.

        :param **state: material state

        :returns: float
        """
        raise NotImplementedError()

    def nu(self, **state):
        """
        Calculate the nu value given the material state.

        :param **state: material state

        :returns: float
        """
        return self.mu(**state) / self.rho(**state)

    def rho(self, **state):
        """
        Calculate the rho value given the material state.

        :param **state: material state

        :returns: float
        """
        raise NotImplementedError()


if __name__ == '__main__':
    import unittest
    from core_test import *
    unittest.main()



# USAGE: you probably want 'setup.py install' - but execute 'setup.py --help'
# for all the details.

# from setuptools import setup, Extension, Command

# from distutils.core import setup

from setuptools import setup

# build the distribution
setup(name='auxi',
      version='0.3.0rc1',
      description='A toolkit to help metallurgical process engineers to '
                  'rapidly do day-to-day calculations.',
      long_description=open('README.md').read(),
      author='Ex Mente (Pty) Ltd',
      author_email='dev@ex-mente.co.za',
      maintainer='Ex Mente (Pty) Ltd',
      maintainer_email='dev@ex-mente.co.za',
      url='https://github.com/Ex-Mente/auxi.0',
      download_url='https://pypi.python.org/pypi/auxi',
      license='LGPL v3.0',
      keywords='metallurgy,chemistry,modelling,simulation,thermochemistry,'
               'engineering,mass balance,energy balance',
      platforms='Ubunutu 14.04,Ubuntu 15.04,Windows 7 (For the stoichiometry '
                'and thermochemistry tools only.)',
      package_dir={'auxi': 'src'},
      packages=['auxi',
                'auxi.core',
                'auxi.modelling',
                'auxi.modelling.process',
                'auxi.modelling.process.materials',
                'auxi.modelling.financial',
                'auxi.modelling.business',
                'auxi.tools',
                'auxi.tools.chemistry',
                'auxi.tools.materialphysicalproperties',
                'auxi.tools.transportphenomena',
                'auxi.tools.transportphenomena.heattransfer',
                'auxi.examples'],
      package_data={'auxi.tools.chemistry': [r'data/*.json',
                                             r'data/rao/*.json',
                                             r'data/nist/*.json'],
                    'auxi.modelling.process.materials': [r'data/*.txt'],
                    'auxi.tools.materialphysicalproperties': [r'data/*.json',
                                                              r'data/*.csv'],
                    'auxi': [r'../*.md', r'doc/*.pdf']},
      install_requires=['jsonpickle', 'tabulate', 'enum34', 'python-dateutil',
                        'cerberus', 'pandas', 'bibtexparser']
      )
